Require Import computable_total_ordering.
Require Import runtime.
Require Import maps.
Require domains.widening_chains.

Module Make (Domain : ORDERED_SET)
  : ABSTRACT_MAP with Module Domain:=Domain.
Module Func_maps := Functional_maps(Domain).

Module Domain := Domain.
Definition domain := Domain.t.
Notation domain_le := Domain.le.
Notation domain_le_refl := Domain.le_refl.
Notation domain_le_trans := Domain.le_trans.
Notation compare := Domain.compare.

Require Import Omega.

Require Import Sumbool.
Implicit Arguments sumbool_and [ A B C D ].
Implicit Arguments sumbool_or [ A B C D ].

Hint Immediate domain_le_trans : emap.
Hint Resolve domain_le_refl : emap.

Lemma le_then_LESS : forall x y : domain,
  (domain_le x y) ->
  ~(domain_le y x) ->
  forall S : Set, forall a b c : S,
  (match (compare x y) with
   | LESS _ _ => a
   | EQUAL _ _ => b
   | GREATER _ _ => c
   end) = a.
Proof.
  intros.
  destruct (compare x y); [ trivial | contradiction | contradiction ].
Qed.

Lemma le_then_EQUAL : forall x y : domain,
  (domain_le x y) ->
  (domain_le y x) ->
  forall S : Set, forall a b c : S,
  (match (compare x y) with
   | LESS _ _ => a
   | EQUAL _ _ => b
   | GREATER _ _ => c
   end) = b.
Proof.
  intros.
  destruct (compare x y); [ contradiction | trivial | contradiction ].
Qed.

Lemma le_then_GREATER : forall x y : domain,
  (domain_le y x) ->
  ~(domain_le x y) ->
  forall S : Set, forall a b c : S,
  (match (compare x y) with
   | LESS _ _ => a
   | EQUAL _ _ => b
   | GREATER _ _ => c
   end) = c.
Proof.
  intros.
  destruct (compare x y); [ contradiction | contradiction | trivial].
Qed.

Lemma le_refl_EQUAL : forall x : domain,
  forall S : Set, forall a b c : S,
  (match (compare x x) with
   | LESS _ _ => a
   | EQUAL _ _ => b
   | GREATER _ _ => c
   end) = b.
Proof.
  intros.
  apply le_then_EQUAL; auto with emap.
Qed.

Ltac solve_emap := solve
         [ apply False_ind; eauto with emap | eauto with emap ] ||
          idtac.

Ltac compare_cases x y := destruct (compare x y); simpl; auto.

Ltac rewrite_LESS :=
  match goal with H1 : (domain_le ?a ?b), H2 : ~(domain_le ?b ?a) |- context [compare ?a ?b] => rewrite (le_then_LESS a b) ; [ idtac | exact H1 | exact H2] end.
Ltac rewrite_EQUAL :=
  match goal with H1 : (domain_le ?a ?b), H2 : (domain_le ?b ?a) |- context [compare ?a ?b] => rewrite (le_then_EQUAL a b) ; [ idtac | exact H1 | exact H2] end.
Ltac rewrite_GREATER :=
  match goal with H1 : (domain_le ?b ?a), H2 : ~(domain_le ?a ?b) |- context [compare ?a ?b] => rewrite (le_then_GREATER a b) ; [ idtac | exact H1 | exact H2] end.

Ltac order_rewrites :=
  repeat first [ rewrite_LESS | rewrite_EQUAL | rewrite_GREATER ].

Ltac compare_cases1 :=
  match goal with |- context [ compare ?a ?b ] => compare_cases a b end.

Ltac compare_cases_auto :=
  repeat ( compare_cases1 ; solve_emap ; order_rewrites).

Ltac absurd_arith := apply False_ind; eauto with arith.
Ltac is_nonempty v := destruct v; [absurd_arith | simpl].

Ltac phys_eq_cases :=
    match goal with |- context [physical_equality ?a ?b] => destruct (physical_equality a b) end.

Require Import Arith.
Require Import Max.

Require assoc_list.
Module alist := assoc_list.Make(Domain).

Section Emap_set.
  Variable codomain : Set.

  Inductive emap_set : Set :=
    Empty_emap : emap_set
  | Node_emap : emap_set -> domain -> codomain -> emap_set -> nat -> emap_set.

  Definition height (m : emap_set) : nat :=
    match m with
      Empty_emap => O
    | (Node_emap _ _ _ _ h) => h
    end.
 
  Inductive emap_bounded_down (bound : domain) : emap_set -> Prop :=
    Empty_emap_bounded_down : (emap_bounded_down bound Empty_emap)
  | Node_emap_bounded_down :
      forall l : emap_set,
      forall d : domain,
      forall c : codomain,
      forall r : emap_set,
      forall h : nat,
      (emap_bounded_down bound l) ->
      (domain_le bound d) -> ~(domain_le d bound) ->
      (emap_bounded_down bound r) ->
      (emap_bounded_down bound (Node_emap l d c r h)).

  Inductive emap_bounded_up (bound : domain) : emap_set -> Prop :=
    Empty_emap_bounded_up : (emap_bounded_up bound Empty_emap)
  | Node_emap_bounded_up :
      forall l : emap_set,
      forall d : domain,
      forall c : codomain,
      forall r : emap_set,
      forall h : nat,
      (emap_bounded_up bound l) ->
      (domain_le d bound) -> ~(domain_le bound d) ->
      (emap_bounded_up bound r) ->
      (emap_bounded_up bound (Node_emap l d c r h)).

  Inductive emap_wo : emap_set -> Prop :=
    Empty_emap_wo : (emap_wo Empty_emap)
  | Node_emap_wo :
      forall l : emap_set,
      forall d : domain,
      forall c : codomain,
      forall r : emap_set, 
      forall h : nat,
      (emap_wo l) ->
      (emap_bounded_up d l) ->
      (emap_bounded_down d r) ->

      (emap_wo r) ->
      h = (S (max (height l) (height r))) ->
      (emap_wo (Node_emap l d c r h)).

  Hint Resolve
    Empty_emap_bounded_down
    Node_emap_bounded_down
    Empty_emap_bounded_up
    Node_emap_bounded_up
    Empty_emap_wo
    Node_emap_wo : emap.

  Ltac inversions_all := repeat
    match goal with
    | H : emap_bounded_up _ (Node_emap _ _ _ _ _) |- _ => inversion_clear H
    | H : emap_bounded_down _ (Node_emap _ _ _ _ _) |- _ => inversion_clear H
    | H : emap_wo (Node_emap _ _ _ _ _) |- _ => inversion_clear H
    end.

 
  Lemma emap_bounded_down_trans :
    forall m : emap_set,
    forall bound1 bound2 : domain,
    (domain_le bound1 bound2) ->
    (emap_bounded_down bound2 m) ->
    (emap_bounded_down bound1 m).
  Proof.
    induction m; intros bound1 bound2 Horder Hbound; inversion_clear Hbound;
    eauto 10 with emap.
  Qed.
  Hint Resolve emap_bounded_down_trans : emap.

  Lemma emap_bounded_up_trans :
    forall m : emap_set,
    forall bound1 bound2 : domain,
    (domain_le bound1 bound2) ->
    (emap_bounded_up bound1 m) ->
    (emap_bounded_up bound2 m).
  Proof.
    induction m; intros bound1 bound2 Horder Hbound; inversion_clear Hbound;
    eauto 10 with emap.
  Qed.
  Hint Resolve emap_bounded_up_trans : emap.

  Definition t := { x : emap_set | (emap_wo x) }.

  Fixpoint emap_find (lookup : domain) (m : emap_set) {struct m} :
    (option codomain) :=
    match m with
      Empty_emap => None
    | (Node_emap l d c r h) =>
      match (compare lookup d) with
        LESS _ _ => emap_find lookup l
      | EQUAL _ _ => Some c
      | GREATER _ _ => emap_find lookup r
      end
    end.

  Lemma emap_find_bounded_up :
    forall m : emap_set,
    forall lookup : domain,
    (emap_bounded_up lookup m) ->
    (emap_find lookup m) = None.
  Proof.
    induction m; simpl in |- *; intros; auto with emap.
    inversions_all; compare_cases lookup d; solve_emap.
  Qed.

  Lemma emap_find_bounded_down :
    forall m : emap_set,
    forall lookup : domain,
    (emap_bounded_down lookup m) ->
    (emap_find lookup m) = None.
  Proof.
    induction m; simpl in |- *; intros; auto with emap.
    inversions_all; compare_cases lookup d; solve_emap.
  Qed.

  Hint Resolve emap_find_bounded_up emap_find_bounded_down : emap.

  Definition find (lookup : domain) (m : t) :=
    emap_find lookup (proj1_sig m).

  Definition create (l:emap_set) (d:domain) (c:codomain) (r:emap_set) :=
    (Node_emap l d c r (S (max (height l) (height r)))).

  Lemma create_wo :
    forall l : emap_set,
    forall d : domain,
    forall c : codomain,
    forall r : emap_set,
    (emap_wo l) ->
    (emap_bounded_up d l) ->
    (emap_bounded_down d r) ->
    (emap_wo r) ->
    (emap_wo (create l d c r)).
  Proof.
    unfold create in |- *; simpl in |- *; auto with emap.
  Qed.

  Lemma create_bounded_down :
    forall bound : domain,
    forall l : emap_set,
    forall d : domain,
    forall c : codomain,
    forall r : emap_set,
    (emap_bounded_down bound l) ->
    (domain_le bound d) ->
    ~(domain_le d bound) ->
    (emap_bounded_down bound r) ->
    (emap_bounded_down bound (create l d c r)).
  Proof.
    unfold create in |- *; simpl in |- *; auto with emap.
  Qed.

  Lemma create_bounded_up :
    forall bound : domain,
    forall l : emap_set,
    forall d : domain,
    forall c : codomain,
    forall r : emap_set,
    (emap_bounded_up bound l) ->
    (domain_le d bound) ->
    ~(domain_le bound d) ->
    (emap_bounded_up bound r) ->
    (emap_bounded_up bound (create l d c r)).
  Proof.
    unfold create in |- *; simpl in |- *; auto with emap.
  Qed.

  Hint Resolve create_wo create_bounded_up create_bounded_down : emap.

  Lemma zero_is_not_greater_than_n :
    forall n : nat, ~(0 > n).
  Proof.
    intro; omega.
  Qed.

  Lemma zero_is_not_greater_than_n2 :
    forall n : nat, (0 > n) -> False.
  Proof.
    intro; omega.
  Qed.

  Hint Resolve zero_is_not_greater_than_n2 : arith.

  Definition map_height_is_nonempty
    (X : Set)
    (m : emap_set)
    (f : emap_set -> domain -> codomain -> emap_set -> X)
    (n : nat)
    (Hheight : (height m) > n) : X.
  Proof.
    unfold height in |- *.
    destruct m as [| l d c r h]; intros.
     apply False_rec.
       exact (zero_is_not_greater_than_n n Hheight).
     exact (f l d c r).
  Defined.

  Definition bal (l:emap_set) (d:domain) (c:codomain) (r:emap_set)
   : emap_set :=
    let hl := height l in
    let hr := height r in
    match (le_gt_dec hl (hr + 2)%nat) with
    | right Hhl2 => (* hl > (hr + 2) *)
      map_height_is_nonempty emap_set l
      (fun ll ld lc lr =>
        match (le_gt_dec (height lr) (height ll)) with
        | left _ => create ll ld lc (create lr d c r)
        | right Hrl0 =>
          map_height_is_nonempty emap_set lr
          (fun lrl lrd lrc lrr =>
	    create (create ll ld lc lrl) lrd lrc (create lrr d c r))
          (height ll) Hrl0
        end
      ) (hr + 2)%nat Hhl2
    | left _ =>
      match (le_gt_dec hr (hl + 2)%nat) with
      | right Hhr2 => (* hr > (hl + 2) *)
        map_height_is_nonempty emap_set r
        (fun rl rd rc rr =>
          match (le_gt_dec (height rl) (height rr)) with
          | left _ => create (create l d c rl) rd rc rr
          | right Hrl0 =>
            map_height_is_nonempty emap_set rl
            (fun rll rld rlc rlr =>
              create (create l d c rll) rld rlc (create rlr rd rc rr))
            (height rr) Hrl0
          end
        ) (hl + 2)%nat Hhr2
      | left _ => (Node_emap l d c r (S (max hl hr)))
      end
    end.

  Lemma bal_bounded_down :
    forall bound : domain,
    forall l : emap_set,
    forall d : domain,
    forall c : codomain,
    forall r : emap_set,
    (emap_bounded_down bound l) ->
    (domain_le bound d) ->
    ~(domain_le d bound) ->
    (emap_bounded_down bound r) ->
    (emap_bounded_down bound (bal l d c r)).
  Proof.
    intros.
    unfold bal in |- *.
    destruct (le_gt_dec (height l) (height r + 2)).
     destruct (le_gt_dec (height r) (height l + 2)).
      auto with emap.
      unfold map_height_is_nonempty in |- *.
        destruct r.
       simpl in g.
         absurd_arith.
       destruct (le_gt_dec (height r1) (height r2)).
        inversion_clear H2.
          auto with emap.
        destruct r1.
         absurd_arith.
         inversion_clear H2.
           inversion_clear H3.
           auto with emap.
     unfold map_height_is_nonempty in |- *.
       destruct l.
      absurd_arith.
      destruct (le_gt_dec (height l2) (height l1)).
       inversion_clear H.
         auto with emap.
       destruct l2.
        absurd_arith.
        inversion_clear H.
          inversion_clear H6.
          auto with emap.
  Qed.

  Lemma bal_bounded_up :
    forall bound : domain,
    forall l : emap_set,
    forall d : domain,
    forall c : codomain,
    forall r : emap_set,
    (emap_bounded_up bound l) ->
    (domain_le d bound) ->
    ~(domain_le bound d) ->
    (emap_bounded_up bound r) ->
    (emap_bounded_up bound (bal l d c r)).
  Proof.
    intros.
    unfold bal in |- *.
    destruct (le_gt_dec (height l) (height r + 2)).
     destruct (le_gt_dec (height r) (height l + 2)).
      auto with emap.
      unfold map_height_is_nonempty in |- *.
        destruct r.
       simpl in g.
         absurd_arith.
       destruct (le_gt_dec (height r1) (height r2)).
        inversion_clear H2.
          auto with emap.
        destruct r1.
         absurd_arith.
         inversion_clear H2.
           inversion_clear H3.
           auto with emap.
     unfold map_height_is_nonempty in |- *.
       destruct l.
      absurd_arith.
      destruct (le_gt_dec (height l2) (height l1)).
       inversion_clear H.
         auto with emap.
       destruct l2.
        absurd_arith.
        inversion_clear H.
          inversion_clear H6.
          auto with emap.
  Qed.

  Hint Resolve bal_bounded_up bal_bounded_down : emap.

  Lemma bal_wo :
    forall l : emap_set,
    forall d : domain,
    forall c : codomain,
    forall r : emap_set,
    (emap_wo l) ->
    (emap_bounded_up d l) ->
    (emap_bounded_down d r) ->
    (emap_wo r) ->
    (emap_wo (bal l d c r)).
  Proof.
    intros l d c r Hlwo Hboundup Hbounddown Hrwo.
    unfold bal in |- *.
    destruct (le_gt_dec (height l) (height r + 2)).
     destruct (le_gt_dec (height r) (height l + 2)).
      auto with emap.
      unfold map_height_is_nonempty in |- *.
        destruct r.
       absurd_arith.
       inversion_clear Hbounddown.
         inversion_clear Hrwo.
         destruct (le_gt_dec (height r1) (height r2)).
        eauto with emap.
        destruct r1.
         absurd_arith.
         inversion_clear H.
           apply create_wo.
          apply create_wo; auto with emap.
            inversion H3; assumption.
          apply create_bounded_up; eauto with emap.
            inversion H3; assumption.
          inversion_clear H4.
            apply create_bounded_down; auto with emap.
           inversion_clear H3; assumption.
           eauto with emap.
          apply create_wo.
           inversion_clear H3; assumption.
           inversion_clear H4; assumption.
           inversion_clear H4; assumption.
           assumption.
     unfold map_height_is_nonempty in |- *.
       destruct l.
      absurd_arith.
      inversion_clear Hlwo.
        destruct (le_gt_dec (height l2) (height l1)).
       apply create_wo; auto with emap.
        inversion_clear Hboundup.
          apply create_bounded_down; auto with emap.
          eauto with emap.
        apply create_wo; auto with emap.
          inversion_clear Hboundup; assumption.
       destruct l2.
        absurd_arith.
        inversion_clear H1.
          inversion_clear H2.
          inversion_clear Hboundup.
          inversion_clear H14.
          apply create_wo; auto with emap.
         apply create_bounded_up; auto with emap.
           eauto with emap.
         apply create_bounded_down; eauto with emap.
  Qed.

  Hint Resolve bal_wo : emap.


  Lemma bal_correct :
    forall l : emap_set,
    forall d : domain,
    forall c : codomain,
    forall r : emap_set,
    (emap_wo l) ->
    (emap_bounded_up d l) ->
    (emap_bounded_down d r) ->
    (emap_wo r) ->
    forall lookup : domain,
    (emap_find lookup (bal l d c r)) =
    (emap_find lookup (create l d c r)). 
  Proof.
    intros.
    unfold bal in |- *.
    destruct (le_gt_dec (height l) (height r + 2)).
     destruct (le_gt_dec (height r) (height l + 2)).
      simpl in |- *.
        trivial.
      unfold map_height_is_nonempty in |- *.
        destruct r.
       absurd_arith.
       destruct (le_gt_dec (height r1) (height r2)).
        inversion_clear H1.
          simpl in |- *.
          compare_cases lookup d0; compare_cases lookup d; solve_emap.
        destruct r1.
         absurd_arith.
         inversions_all.
           simpl in |- *.
           compare_cases lookup d; compare_cases lookup d0;
            compare_cases lookup d1; solve_emap.
     is_nonempty l.
       destruct (le_gt_dec (height l2) (height l1)); simpl in |- *.
      inversions_all.
        compare_cases lookup d; compare_cases lookup d0; solve_emap.
      is_nonempty l2.
        inversions_all.
        compare_cases lookup d; compare_cases lookup d0; compare_cases lookup d1;
         solve_emap.
  Qed.

  Definition empty : unit -> t.
  Proof.
    intro dummy.
    exists Empty_emap.
    auto with emap.
  Defined.

  Definition is_empty (m : t) : bool :=
  match (proj1_sig m) with
    Empty_emap => true
  | Node_emap _ _ _ _ _ => false
  end.

  Fixpoint emap_add (key : domain) (img : codomain) (m : emap_set)
    {struct m} : emap_set :=
    match m with
    | Empty_emap => Node_emap Empty_emap key img Empty_emap 1%nat
    | (Node_emap l d c r h) =>
      match compare key d with
      | EQUAL _ _ =>
          match (physical_equality img c) with
          | left equals => m
          | right _ => (Node_emap l key img r h)
          end
      | LESS _ _ =>
          let lprime := (emap_add key img l) in
          match (physical_equality l lprime) with
          | left equals => m
          | right _ => bal lprime d c r
          end
      | GREATER _ _ =>
          let rprime := (emap_add key img r) in
          match (physical_equality r rprime) with
          | left equals => m 
          | right _ => bal l d c rprime
          end
      end
  end.

  Lemma add_bounded_up :
    forall bound : domain,
    forall key : domain,
    (domain_le key bound) -> ~(domain_le bound key) ->
    forall img : codomain,
    forall m : emap_set,
    (emap_bounded_up bound m) ->
    (emap_bounded_up bound (emap_add key img m)).
  Proof.
    induction m; intro Hbound; inversion Hbound; simpl in |- *.
     auto with emap.
     compare_cases key d.
      destruct (physical_equality m1 (emap_add key img m1)).
       rewrite e.
         auto with emap.
       auto with emap.
      destruct (physical_equality img c); auto with emap.
      destruct (physical_equality m2 (emap_add key img m2)).
       rewrite e.
         auto with emap.
       auto with emap.
  Qed.

  Lemma add_bounded_down :
    forall bound : domain,
    forall key : domain,
    (domain_le bound key) -> ~(domain_le key bound) ->
    forall img : codomain,
    forall m : emap_set,
    (emap_bounded_down bound m) ->
    (emap_bounded_down bound (emap_add key img m)).
  Proof.
    induction m; intro Hbound; inversion Hbound; simpl in |- *.
     auto with emap.
     compare_cases key d.
      destruct (physical_equality m1 (emap_add key img m1)).
       rewrite e.
         auto with emap.
       auto with emap.
      destruct (physical_equality img c); auto with emap.
      destruct (physical_equality m2 (emap_add key img m2)).
       rewrite e.
         auto with emap.
       auto with emap.
  Qed.

  Hint Resolve add_bounded_up add_bounded_down : emap.

  Lemma add_wo :
    forall key : domain, 
    forall img : codomain,
    forall m : emap_set,
    forall Hmwo : (emap_wo m),
    (emap_wo (emap_add key img m)).
  Proof.
    induction m; intro Hwo; inversion_clear Hwo; simpl in |- *.
     auto with emap.
     compare_cases key d.
      destruct (physical_equality m1 (emap_add key img m1)); auto with emap.
      destruct (physical_equality img c); auto with emap.
        eauto with emap.
      destruct (physical_equality m2 (emap_add key img m2)); auto with emap.
  Qed.

  Hint Resolve add_wo add_bounded_up add_bounded_down : emap.


  Fixpoint prepend_emap_to_assoc_list (m : emap_set)
    (tail : (alist.alist_set codomain)) {struct m} :
    (alist.alist_set codomain) :=
    match m with
    | Empty_emap => tail
    | (Node_emap l d c r _) =>
        (prepend_emap_to_assoc_list l
          (alist.Cons_alist codomain d c (prepend_emap_to_assoc_list r tail)))
    end.

  Lemma prepend_emap_to_assoc_list_bounded_down :
    forall bound : domain,
    forall m :emap_set,
    forall tail : (alist.alist_set codomain),
    (emap_bounded_down bound m) ->
    (alist.alist_bounded codomain bound tail) ->
    (alist.alist_bounded codomain bound
      (prepend_emap_to_assoc_list m tail)).
  Proof.
    induction m; simpl in |- *; intros; inversions_all; auto with emap alist.
  Qed.

  Hint Resolve prepend_emap_to_assoc_list_bounded_down : emap.

  Section alist_bounds.
    Variable tail : (alist.alist_set codomain).
    Inductive alist_bounded_by_emap : emap_set -> Prop :=
    | Alist_bounded_by_empty_emap : (alist_bounded_by_emap Empty_emap)
    | Alist_bounded_by_node_emap :
        forall l : emap_set,
        forall d : domain,
        forall c : codomain,
        forall r : emap_set,
        forall h : nat,
        (alist.alist_bounded codomain d tail) ->
        (alist_bounded_by_emap r) ->
        (alist_bounded_by_emap (Node_emap l d c r h)).
  End alist_bounds.

  Lemma cons_alist_bounded_by_emap :
    forall m : emap_set,
    forall d : domain,
    forall c : codomain,
    forall tail : (alist.alist_set codomain),
    (emap_bounded_up d m) ->
    (alist.alist_bounded codomain d tail) ->
    (alist_bounded_by_emap tail m) ->
    (alist_bounded_by_emap (alist.Cons_alist codomain d c tail) m).
  Proof.
    induction m; simpl in |- *; intros d1 c1 tail Hbound1 Hbound2 Hbound3.
     apply Alist_bounded_by_empty_emap.
     inversion_clear Hbound3.
       inversion_clear Hbound1.
       apply Alist_bounded_by_node_emap.
      apply alist.Cons_alist_bounded; trivial.
      apply IHm2; assumption.
  Qed.

  Hint Resolve cons_alist_bounded_by_emap : emap.

  Lemma alist_bounded_by_prepend_emap :
    forall m1 : emap_set,
    forall d : domain,
    forall m2 : emap_set,
    forall tail : (alist.alist_set codomain),
    (emap_bounded_up d m1) ->
    (emap_bounded_down d m2) ->
    (alist_bounded_by_emap tail m2) ->
    (alist.alist_bounded codomain d tail) ->
    (alist_bounded_by_emap (prepend_emap_to_assoc_list m2 tail) m1).
  Proof.
    induction m1; intros md m2 tail Hbound1 Hbound2 Hbound3 Hbound4.
     apply Alist_bounded_by_empty_emap.
     apply Alist_bounded_by_node_emap.
      inversion_clear Hbound1.
    apply prepend_emap_to_assoc_list_bounded_down.
       apply emap_bounded_down_trans with (bound2 := md); assumption.
       apply alist.alist_le_bounded with (bound2 := md); assumption.
      inversion_clear Hbound1.
    apply IHm1_2 with (d := md); assumption.
  Qed.

  Lemma prepend_emap_to_assoc_list_wo :
    forall m :emap_set,
    forall tail : (alist.alist_set codomain),
    (emap_wo m) ->
    (alist_bounded_by_emap tail m) ->
    (alist.alist_wo codomain tail) ->
    (alist.alist_wo codomain (prepend_emap_to_assoc_list m tail)).
  Proof.
    induction m; simpl in |- *; intros tail Hm_wo Hbound1 Htail_wo.
     assumption.
     inversion_clear Hm_wo.
       inversion_clear Hbound1.
       apply IHm1.
      assumption.
      apply cons_alist_bounded_by_emap; auto with emap.
    apply alist_bounded_by_prepend_emap with (d := d); trivial.
      auto with alist emap.
  Qed.

  Lemma prepend_emap_to_assoc_list_correct :
    forall m :emap_set,
    forall tail : (alist.alist_set codomain),
    (emap_wo m) ->
    (alist_bounded_by_emap tail m) ->
    (alist.alist_wo codomain tail) ->
    forall lookup : domain,
    (alist.alist_find codomain
       (prepend_emap_to_assoc_list m tail) lookup) =
    (match (emap_find lookup m) with
     | Some x => Some x
     | None => (alist.alist_find codomain tail lookup)
     end).
  Proof.
    induction m; intros tail Hm_wo Hbound Htail_wo lookup; simpl in |- *.
     trivial.
     inversions_all.
       rewrite IHm1; trivial.
      compare_cases lookup d; order_rewrites.
       rewrite alist.alist_bounded_find_empty.
    trivial.
    inversion Hbound.
      apply alist.alist_le_bounded with (bound2 := d); assumption.
       rewrite emap_find_bounded_up.
    trivial.
    eauto with emap.
       rewrite (emap_find_bounded_up m1).
    inversion Hbound.
      apply IHm2; trivial.
    eauto with emap.
      apply cons_alist_bounded_by_emap.
       assumption.
       inversion Hbound.
     auto with emap.
       inversion Hbound.
     apply alist_bounded_by_prepend_emap with (d := d); trivial.
      inversion Hbound.
    apply alist.Cons_alist_wo.
       apply prepend_emap_to_assoc_list_wo; trivial.
       auto with emap.
  Qed.

  Definition emap_to_assoc_list (m : emap_set) :=
    prepend_emap_to_assoc_list m (alist.Empty_alist codomain).

  Lemma alist_bounded_by_empty_emap :
    forall m : emap_set,
    (alist_bounded_by_emap (alist.Empty_alist codomain) m).
  Proof.
    induction m; simpl in |- *.
     apply Alist_bounded_by_empty_emap.
     inversions_all.
       apply Alist_bounded_by_node_emap; auto with emap alist.
  Qed.

  Lemma emap_to_assoc_list_wo :
    forall m : emap_set,
    (emap_wo m) ->
    (alist.alist_wo codomain (emap_to_assoc_list m)).
  Proof.
    unfold emap_to_assoc_list in |- *.
    intros.
    apply prepend_emap_to_assoc_list_wo.
     trivial.
     apply alist_bounded_by_empty_emap.
     apply alist.Empty_alist_wo.
  Qed.

  Hint Resolve emap_to_assoc_list_wo : emap.

  Lemma emap_to_assoc_list_correct :
    forall m : emap_set,
    (emap_wo m) ->
    forall lookup : domain,
    (emap_find lookup m) =
    (alist.alist_find codomain (emap_to_assoc_list m) lookup).
  Proof.
    intros.
    unfold emap_to_assoc_list in |- *.
    rewrite prepend_emap_to_assoc_list_correct; auto with emap alist.
     simpl in |- *.
       destruct (emap_find lookup m); trivial.
     apply alist_bounded_by_empty_emap.
  Qed.

  Definition to_assoc_list (m : t) : (alist.t codomain).
    intro.
    destruct m as (m, Hm).
    exists (emap_to_assoc_list m).
    apply emap_to_assoc_list_wo; assumption.
  Defined.

  Theorem to_assoc_list_correct:
    forall m : t,
    forall lookup : domain,
    (find lookup m) =
    (alist.find codomain lookup (to_assoc_list m)).
  Proof.
    intros.
    destruct m as (m, Hm).
    unfold find, alist.find, to_assoc_list in |- *; simpl in |- *.
    apply emap_to_assoc_list_correct.
    assumption.
  Qed.

  Definition func_add := (Func_maps.func_add codomain).

  Definition add (key : domain) (img : codomain) (m : t) : t.
  Proof.
    intros.
    destruct m as (m, Hmwo).
    exists (emap_add key img m).
    apply add_wo; assumption.
  Defined.

  Lemma emap_add_correct :
    forall key : domain,
    forall img : codomain,
    forall m : emap_set,
    (emap_wo m) ->
    forall lookup : domain,
    (func_add key img (fun lookup1 => emap_find lookup1 m) lookup) =
    (emap_find lookup (emap_add key img m)).
  Proof.
    unfold func_add in |- *.
    unfold Func_maps.func_add in |- *.
    induction m; simpl in |- *.
     trivial.
     intros.
       inversions_all.
       compare_cases lookup key; compare_cases lookup d; compare_cases key d;
        solve_emap; phys_eq_cases; simpl in |- *; order_rewrites;
        try solve
         [ auto with emap
         | rewrite bal_correct; auto with emap; simpl in |- *; order_rewrites;
            solve
            [ auto with emap
            | rewrite <- IHm1; order_rewrites; auto with emap
            | rewrite <- IHm2; order_rewrites; auto with emap ] ].
      rewrite e.
      rewrite <- IHm1; order_rewrites; auto with emap.
      rewrite e; trivial.
      rewrite e.
      rewrite <- IHm2; order_rewrites; auto with emap.
  Qed.

  Theorem add_correct :
    forall key : domain,
    forall img : codomain,
    forall m : t,
    forall lookup : domain,
    (find lookup (add key img m)) =
    (func_add key img (fun lookup1 => find lookup1 m) lookup).
  Proof.
    destruct m as (m, Hmwo).
    intro.
    unfold find, add in |- *; simpl in |- *.
    symmetry.
    apply emap_add_correct; assumption.
  Qed.

  Fixpoint min_binding
    (l : emap_set) (d : domain) (c : codomain)
    {struct l} : (prod domain codomain) :=
  match l with
  | Empty_emap => (d, c)
  | (Node_emap l2 x2 d2 _ _) => (min_binding l2 x2 d2)
  end.

  Fixpoint remove_min_binding
   (l : emap_set) (d : domain) (c : codomain) (r : emap_set)
   {struct l} : emap_set :=
  match l with
  | Empty_emap => r
  | (Node_emap l2 d2 c2 r2 _) =>
    (bal (remove_min_binding l2 d2 c2 r2) d c r)
  end.

  Lemma remove_min_binding_bounded_up : 
   forall bound : domain,
   forall l : emap_set,
   forall d : domain,
   forall c : codomain,
   forall r : emap_set,
   (emap_bounded_up bound l) ->
   (domain_le d bound) ->
   ~(domain_le bound d) ->
   (emap_bounded_up bound r) ->
   (emap_bounded_up bound (remove_min_binding l d c r)).
  Proof.
    induction l; simpl in |- *; intros; inversions_all; auto with emap.
  Qed.

  Hint Resolve remove_min_binding_bounded_up : emap.

  Lemma remove_min_binding_bounded_down : 
   forall bound : domain,
   forall l : emap_set,
   forall d : domain,
   forall c : codomain,
   forall r : emap_set,
   (emap_bounded_down bound l) ->
   (domain_le bound d) ->
   ~(domain_le d bound) ->
   (emap_bounded_down bound r) ->
   (emap_bounded_down bound (remove_min_binding l d c r)).
  Proof.
    induction l; simpl in |- *; intros; inversions_all; auto with emap.
  Qed.

  Hint Resolve remove_min_binding_bounded_down : emap.

  Lemma remove_min_binding_wo : 
   forall l : emap_set,
   forall d : domain,
   forall c : codomain,
   forall r : emap_set,
   (emap_wo l) ->
   (emap_bounded_up d l) ->
   (emap_bounded_down d r) ->
   (emap_wo r) ->
   (emap_wo (remove_min_binding l d c r)).
  Proof.
    induction l; simpl in |- *; intros; inversions_all; auto with emap.
  Qed.

  Hint Resolve remove_min_binding_wo : emap.

  Lemma remove_min_binding_correct :
    forall l : emap_set,
    forall d : domain,
    forall c : codomain,
    forall r : emap_set,
    (emap_wo l) ->
    (emap_bounded_up d l) ->
    (emap_bounded_down d r) ->
    (emap_wo r) ->
    forall lookup,
    (domain_le (fst (min_binding l d c)) lookup) ->
    ~(domain_le lookup (fst (min_binding l d c))) ->
    (emap_find lookup (remove_min_binding l d c r)) =
    (emap_find lookup (create l d c r)).
  Proof.
    induction l; simpl in |- *; intros; inversions_all; auto with emap.
    order_rewrites; trivial.
    rewrite bal_correct; auto with emap; simpl in |- *.
    compare_cases_auto; rewrite IHl1; simpl in |- *; order_rewrites;
     auto with emap.
  Qed.

  Hint Resolve remove_min_binding_correct : emap.

  Lemma min_binding_bounded_up1 :
    forall bound : domain,
    forall l : emap_set,
    forall d : domain,
    forall c : codomain,
    (emap_bounded_up bound l) ->
    (domain_le d bound) ->
    ~(domain_le bound d) ->
    (domain_le (fst (min_binding l d c)) bound).
  Proof.
    induction l; simpl in |- *; intros; inversions_all; auto with emap.
  Qed.

  Lemma min_binding_bounded_down1 :
    forall bound : domain,
    forall l : emap_set,
    forall d : domain,
    forall c : codomain,
    (emap_bounded_down bound l) ->
    (domain_le bound d) ->
    ~(domain_le d bound) ->
    (domain_le bound (fst (min_binding l d c))).
  Proof.
    induction l; simpl in |- *; intros; inversions_all; auto with emap.
  Qed.

  Lemma min_binding_bounded_up2 :
    forall bound : domain,
    forall l : emap_set,
    forall d : domain,
    forall c : codomain,
    (emap_bounded_up bound l) ->
    (domain_le d bound) ->
    ~(domain_le bound d) ->
    ~(domain_le bound (fst (min_binding l d c))).
  Proof.
    induction l; simpl in |- *; intros; inversions_all; auto with emap.
  Qed.

  Lemma min_binding_bounded_down2 :
    forall bound : domain,
    forall l : emap_set,
    forall d : domain,
    forall c : codomain,
    (emap_bounded_down bound l) ->
    (domain_le bound d) ->
    ~(domain_le d bound) ->
    ~(domain_le(fst (min_binding l d c)) bound).
  Proof.
    induction l; simpl in |- *; intros; inversions_all; auto with emap.
  Qed.

  Lemma min_binding_bounds_down :
    forall l : emap_set,
    forall d : domain,
    forall c : codomain,
    forall r : emap_set,
    forall n : nat,
    (emap_wo l) ->
    (emap_bounded_up d l) ->
    (emap_bounded_down d r) ->
    (emap_wo r) ->
    forall bound : domain,
    (domain_le bound (fst (min_binding l d c))) ->
    ~(domain_le (fst (min_binding l d c)) bound) ->
    (emap_bounded_down bound (Node_emap l d c r n)).
  Proof.
    induction l; simpl in |- *; intros.
     apply Node_emap_bounded_down; auto with emap.
       apply emap_bounded_down_trans with (bound2 := d); assumption.
     apply Node_emap_bounded_down.
      inversions_all; apply IHl1; auto with emap.
      assert (emap_bounded_down bound (Node_emap l1 d c l2 n)).
       inversions_all; apply IHl1; auto with emap.
       inversions_all.
         eauto with emap.
      assert (emap_bounded_down bound (Node_emap l1 d c l2 n)).
       inversions_all; apply IHl1; auto with emap.
       inversions_all; eauto with emap.
      assert (emap_bounded_down bound (Node_emap l1 d c l2 n)).
       inversions_all; apply IHl1; auto with emap.
       inversions_all.
         eauto with emap.
  Qed.

  Lemma min_binding_correct :
    forall l : emap_set,
    forall d : domain,
    forall c : codomain,
    forall r : emap_set,
    forall n : nat,
    (emap_wo l) ->
    (emap_bounded_up d l) ->
    forall lookup : domain,
    (domain_le lookup (fst (min_binding l d c))) ->
    (domain_le (fst (min_binding l d c)) lookup) ->
    (emap_find lookup (Node_emap l d c r n)) = Some (snd (min_binding l d c)).
  Proof.
    induction l; simpl in |- *; intros.
     rewrite le_then_EQUAL; trivial.
     compare_cases lookup d0.
      change
        (emap_find lookup (Node_emap l1 d c l2 n) =
         Some (snd (min_binding l1 d c))) in |- *.
        inversions_all; apply IHl1; auto with emap.
      assert (~ domain_le d0 (fst (min_binding l1 d c))).
       inversions_all; apply min_binding_bounded_up2; auto with emap.
       apply False_ind.
         eauto with emap.
      assert (~ domain_le d0 (fst (min_binding l1 d c))).
       inversions_all; apply min_binding_bounded_up2; auto with emap.
       apply False_ind.
         eauto with emap.
  Qed.

  Hint Resolve min_binding_bounded_up1 min_binding_bounded_down1
    min_binding_bounded_up2 min_binding_bounded_down2 : emap.

  Definition merge (t1 : emap_set) (t2 : emap_set) :=
    match t2 with
    | Empty_emap => t1
    | (Node_emap l2 d2 c2 r2 _) =>
      match t1 with
      | Empty_emap => t2
      | _ => let (d, c) := min_binding l2 d2 c2 in
        bal t1 d c (remove_min_binding l2 d2 c2 r2)
      end
    end.

  Fixpoint emap_remove (key : domain) (m : emap_set) {struct m} : emap_set :=
    match m with
    | Empty_emap => Empty_emap
    | (Node_emap l d c r h) =>
      match (compare key d) with
      | EQUAL _ _ => merge l r
      | LESS _ _ =>
        let lprime := (emap_remove key l) in
        match (physical_equality lprime l) with
        | left _ => m
        | right _ => bal lprime d c r
        end
      | GREATER _ _ =>
        let rprime := (emap_remove key r) in
        match (physical_equality rprime r) with
        | left _ => m
        | right _ => bal l d c rprime
        end
      end
    end.

  Fixpoint emap_remove2 (key : domain) (m : emap_set) {struct m} : emap_set :=
    match m with
    | Empty_emap => Empty_emap
    | (Node_emap l d c r h) =>
      match (compare key d) with
      | EQUAL _ _ => merge l r
      | LESS _ _ => bal (emap_remove2 key l) d c r
      | GREATER _ _ => bal l d c (emap_remove2 key r)
      end
    end.

  Lemma merge_bounded_up :
    forall bound : domain,
    forall m1 m2 : emap_set,
    (emap_bounded_up bound m1) ->
    (emap_bounded_up bound m2) ->
    (emap_bounded_up bound (merge m1 m2)).
  Proof.
    intros bound m1 m2 Hm1 Hm2.
    unfold merge in |- *.
    destruct m2 as [| m2l m2d m2c m2r m2h]; trivial.
    destruct m1 as [| m1l m1d m1c m1r m1h]; trivial.
    set (min2 := min_binding m2l m2d m2c) in |- *.
    replace min2 with (fst min2, snd min2); simpl in |- *.
     replace min2 with (min_binding m2l m2d m2c); trivial.
       inversions_all.
       auto with emap.
     destruct min2; simpl in |- *; trivial.
  Qed.

  Lemma merge_bounded_down :
    forall bound : domain,
    forall m1 m2 : emap_set,
    (emap_bounded_down bound m1) ->
    (emap_bounded_down bound m2) ->
    (emap_bounded_down bound (merge m1 m2)).
  Proof.
    intros bound m1 m2 Hm1 Hm2.
    unfold merge in |- *.
    destruct m2 as [| m2l m2d m2c m2r m2h]; trivial.
    destruct m1 as [| m1l m1d m1c m1r m1h]; trivial.
    set (min2 := min_binding m2l m2d m2c) in |- *.
    replace min2 with (fst min2, snd min2); simpl in |- *.
     replace min2 with (min_binding m2l m2d m2c); trivial.
       inversions_all.
       auto with emap.
     destruct min2; simpl in |- *; trivial.
  Qed.

  Hint Resolve merge_bounded_up merge_bounded_down : emap.

  Lemma remove_min_binding_yields_bounded_down :
    forall l : emap_set,
    forall d : domain,
    forall c : codomain,
    forall r : emap_set,
    (emap_wo l) ->
    (emap_bounded_up d l) ->
    (emap_bounded_down d r) ->
    (emap_wo r) ->
      emap_bounded_down (fst (min_binding l d c))
        (remove_min_binding l d c r).
  Proof.
    induction l as [| ll Hll ld lc lr Hlr lh]; intros; simpl in |- *;
     inversions_all; auto with emap.
    apply bal_bounded_down; auto with emap.
    destruct r as [| rl rd rc rr rh].
     apply Empty_emap_bounded_down.
     inversions_all.
       eauto with emap.
  Qed.

  Hint Resolve remove_min_binding_yields_bounded_down : emap.

  Lemma merge_wo :
    forall m1 : emap_set,
    forall middle : domain,
    forall m2 : emap_set,
    (emap_wo m1) ->
    (emap_bounded_up middle m1) ->
    (emap_bounded_down middle m2) ->
    (emap_wo m2) -> (emap_wo (merge m1 m2)).
  Proof.
    intros m1 middle m2 Hwo1 Hbound1 Hbound2 Hwo2.
    unfold merge in |- *.
    destruct m2 as [| m2l m2d m2c m2r m2h]; trivial.
    destruct m1 as [| m1l m1d m1c m1r m1h]; trivial.
    set (min2 := min_binding m2l m2d m2c) in |- *.
    replace min2 with (fst min2, snd min2); simpl in |- *.
     replace min2 with (min_binding m2l m2d m2c); trivial.
       inversions_all.
       apply bal_wo.
      auto with emap.
      apply emap_bounded_up_trans with (bound1 := middle); auto with emap.
      auto with emap.
      auto with emap.
     destruct min2; simpl in |- *; trivial.
  Qed.

  Hint Resolve merge_wo : emap.

  Lemma merge_correct1 :
    forall m1 : emap_set,
    forall middle : domain,
    forall m2 : emap_set,
    (emap_wo m1) ->
    (emap_bounded_up middle m1) ->
    (emap_bounded_down middle m2) ->
    (emap_wo m2) ->
    forall lookup : domain,
    (domain_le lookup middle) ->
    ~(domain_le middle lookup) ->
    (emap_find lookup (merge m1 m2)) = (emap_find lookup m1).
  Proof.
    intros.
    unfold merge in |- *.
    destruct m2 as [| m2l m2d m2c m2r]; trivial.
    destruct m1 as [| m1l m1d m1c m1r]; simpl in |- *.
     inversions_all; compare_cases_auto; auto with emap.
     set (min2 := min_binding m2l m2d m2c) in |- *.
       replace min2 with (fst min2, snd min2); simpl in |- *.
      replace min2 with (min_binding m2l m2d m2c); trivial.
        rewrite bal_correct; auto with emap.
       assert (domain_le middle (fst (min_binding m2l m2d m2c))).
        apply min_binding_bounded_down1; inversions_all; auto with emap.
        assert (~ domain_le (fst (min_binding m2l m2d m2c)) middle).
         apply min_binding_bounded_down2; inversions_all; auto with emap.
         assert (domain_le lookup (fst (min_binding m2l m2d m2c))).
          eauto with emap.
          assert (~ domain_le (fst (min_binding m2l m2d m2c)) lookup).
           eauto with emap.
           unfold create in |- *; simpl in |- *.
             compare_cases_auto; solve_emap.
       apply emap_bounded_up_trans with (bound1 := middle).
        inversions_all; auto with emap.
        assumption.
       apply remove_min_binding_yields_bounded_down; inversions_all;
        auto with emap.
       inversions_all; auto with emap.
      destruct min2; trivial.
  Qed.

  Lemma merge_correct2 :
    forall m1 : emap_set,
    forall middle : domain,
    forall m2 : emap_set,
    (emap_wo m1) ->
    (emap_bounded_up middle m1) ->
    (emap_bounded_down middle m2) ->
    (emap_wo m2) ->
    forall lookup : domain,
    (domain_le middle lookup) ->
    ~(domain_le lookup middle) ->
    (emap_find lookup (merge m1 m2)) = (emap_find lookup m2).
  Proof.
    intros.
    unfold merge in |- *.
    destruct m2 as [| m2l m2d m2c m2r]; simpl in |- *.
     inversions_all; eauto with emap.
     destruct m1 as [| m1l m1d m1c m1r]; simpl in |- *.
      trivial.
      set (min2 := min_binding m2l m2d m2c) in |- *.
        replace min2 with (fst min2, snd min2); simpl in |- *.
       replace min2 with (min_binding m2l m2d m2c); trivial.
         rewrite bal_correct; auto with emap.
        assert (domain_le middle (fst (min_binding m2l m2d m2c))).
         apply min_binding_bounded_down1; inversions_all; auto with emap.
         assert (~ domain_le (fst (min_binding m2l m2d m2c)) middle).
          apply min_binding_bounded_down2; inversions_all; auto with emap.
          unfold create in |- *; simpl in |- *.
            compare_cases lookup (fst (min_binding m2l m2d m2c)).
           replace
            match compare lookup m1d with
            | LESS _ _ => emap_find lookup m1l
            | EQUAL _ _ => Some m1c
            | GREATER _ _ => emap_find lookup m1r
            end with (None (A:=codomain)).
            replace
             match compare lookup m2d with
             | LESS _ _ => emap_find lookup m2l
             | EQUAL _ _ => Some m2c
             | GREATER _ _ => emap_find lookup m2r
             end with (None (A:=codomain)).
             trivial.
             change (None = emap_find lookup (Node_emap m2l m2d m2c m2r n))
              in |- *.
               symmetry  in |- *; apply emap_find_bounded_down; inversions_all;
                auto with emap.
               apply min_binding_bounds_down; auto with emap.
            symmetry  in |- *.
              change (emap_find lookup (Node_emap m1l m1d m1c m1r n0) = None)
               in |- *.
              apply emap_find_bounded_up; auto with emap.
              apply emap_bounded_up_trans with (bound1 := middle); assumption.
           symmetry  in |- *.
             change
               (emap_find lookup (Node_emap m2l m2d m2c m2r n) =
                Some (snd (min_binding m2l m2d m2c))) in |- *.
             inversions_all; apply min_binding_correct; auto with emap.
           inversions_all; rewrite remove_min_binding_correct; auto with emap.
        inversions_all; eauto with emap.
        inversions_all; eauto with emap.
        inversions_all; auto with emap.
       destruct min2; trivial.
  Qed.

  Lemma merge_correctM :
    forall m1 : emap_set,
    forall middle : domain,
    forall m2 : emap_set,
    (emap_wo m1) ->
    (emap_bounded_up middle m1) ->
    (emap_bounded_down middle m2) ->
    (emap_wo m2) ->
    forall lookup : domain,
    (domain_le middle lookup) ->
    (domain_le lookup middle) ->
    (emap_find lookup (merge m1 m2)) = None.
  Proof.
    intros.
    unfold merge in |- *.
    destruct m2 as [| m2l m2d m2c m2r]; simpl in |- *.
     auto with emap.
       eauto with emap.
     destruct m1.
      assert (emap_bounded_down lookup (Node_emap m2l m2d m2c m2r n)).
       eauto with emap.
       auto with emap.
      set (min2 := min_binding m2l m2d m2c) in |- *.
        replace min2 with (fst min2, snd min2); simpl in |- *.
       replace min2 with (min_binding m2l m2d m2c); trivial.
         inversions_all.
         rewrite bal_correct; auto with emap.
        simpl in |- *.
          assert (domain_le lookup (fst (min_binding m2l m2d m2c))).
         assert (domain_le middle (fst (min_binding m2l m2d m2c))).
          apply min_binding_bounded_down1; assumption.
          eauto with emap.
         assert (~ domain_le (fst (min_binding m2l m2d m2c)) lookup).
          assert (~ domain_le (fst (min_binding m2l m2d m2c)) middle).
           apply min_binding_bounded_down2; assumption.
           eauto with emap.
          order_rewrites.
            replace
             match compare lookup d with
             | LESS _ _ => emap_find lookup m1_1
             | EQUAL _ _ => Some c
             | GREATER _ _ => emap_find lookup m1_2
             end with (emap_find lookup m1_2).
           assert (emap_bounded_up middle m1_2).
            auto with emap.
            apply emap_find_bounded_up.
              apply emap_bounded_up_trans with (bound1 := middle); assumption.
           assert (domain_le d lookup).
            apply domain_le_trans with (y := middle); assumption.
            assert (~ domain_le lookup d).
             intro Z.
               apply H7.
               apply domain_le_trans with (y := lookup); assumption.
             order_rewrites.
               trivial.
        apply emap_bounded_up_trans with (bound1 := middle).
         auto with emap.
         auto with emap.
       destruct min2; trivial.
  Qed.

  Lemma emap_remove2_bounded_up :
    forall bound : domain,
    forall key : domain,
    forall m : emap_set,
    (emap_bounded_up bound m) ->
    (emap_bounded_up bound (emap_remove2 key m)).
  Proof.
    induction m as [| l d c r h]; intros; simpl in |- *; auto with emap.
    inversions_all.
    compare_cases_auto; auto with emap.
  Qed.

  Lemma emap_remove2_bounded_down :
    forall bound : domain,
    forall key : domain,
    forall m : emap_set,
    (emap_bounded_down bound m) ->
    (emap_bounded_down bound (emap_remove2 key m)).
  Proof.
    induction m as [| l d c r h]; intros; simpl in |- *; auto with emap.
    inversions_all.
    compare_cases_auto; auto with emap.
  Qed.

  Hint Resolve emap_remove2_bounded_up emap_remove2_bounded_down : emap.

  Lemma emap_remove2_wo :
    forall key : domain,
    forall m : emap_set,
    (emap_wo m) ->
    (emap_wo (emap_remove2 key m)).
  Proof.
    induction m as [| l d c r h]; intros; simpl in |- *; auto with emap.
    inversions_all.
    compare_cases_auto; auto with emap.
  Qed.

  Hint Resolve emap_remove2_wo : emap.

  Definition func_remove (key : domain)
     (f : (domain -> (option codomain))) (lookup : domain) :
     (option codomain) :=
    match (compare key lookup) with
    | EQUAL _ _ => None
    | _ => (f lookup)
    end.

  Lemma emap_remove2_correct :
    forall key : domain,
    forall m : emap_set,
    (emap_wo m) ->
    forall lookup : domain,
    (func_remove key (fun lookup1 => emap_find lookup1 m) lookup) =
    (emap_find lookup (emap_remove2 key m)).
  Proof.
    unfold func_remove in |- *; induction m; intros; inversions_all;
     compare_cases_auto; auto with emap;
     try solve
      [ rewrite bal_correct; auto with emap; unfold create in |- *; simpl in |- *;
         order_rewrites; auto with emap
      | rewrite bal_correct; auto with emap; unfold create in |- *; simpl in |- *;
         order_rewrites; rewrite <- IHm1; order_rewrites; 
         auto with emap
      | rewrite bal_correct; auto with emap; unfold create in |- *; simpl in |- *;
         order_rewrites; rewrite <- IHm2; order_rewrites; 
         auto with emap ].
     symmetry  in |- *.
       apply merge_correct2 with (middle := d); auto with emap.
     rewrite bal_correct; auto with emap.
       simpl in |- *.
       compare_cases_auto.
       rewrite <- IHm1.
      rewrite le_then_EQUAL; trivial.
      assumption.
     symmetry  in |- *.
       apply merge_correctM with (middle := d); eauto with emap.
     rewrite bal_correct; auto with emap.
       simpl in |- *.
       compare_cases_auto.
       rewrite <- IHm2.
      rewrite le_then_EQUAL; trivial.
      assumption.
     rewrite merge_correct1 with (middle := d); auto with emap.
  Qed.

  Lemma emap_remove_bounded_up :
    forall bound : domain,
    forall key : domain,
    forall m : emap_set,
    (emap_bounded_up bound m) ->
    (emap_bounded_up bound (emap_remove key m)).
  Proof.
    induction m as [| l d c r h]; intros; simpl in |- *; auto with emap.
    inversions_all.
    compare_cases key c.
     phys_eq_cases; auto with emap.
     auto with emap.
     phys_eq_cases; auto with emap.
  Qed.

  Lemma emap_remove_bounded_down :
    forall bound : domain,
    forall key : domain,
    forall m : emap_set,
    (emap_bounded_down bound m) ->
    (emap_bounded_down bound (emap_remove key m)).
  Proof.
    induction m as [| l d c r h]; intros; simpl in |- *; auto with emap.
    inversions_all.
    compare_cases key c.
     phys_eq_cases; auto with emap.
     auto with emap.
     phys_eq_cases; auto with emap.
  Qed.

  Hint Resolve emap_remove_bounded_up emap_remove_bounded_down : emap.

  Lemma emap_remove_wo :
    forall key : domain,
    forall m : emap_set,
    (emap_wo m) ->
    (emap_wo (emap_remove key m)).
  Proof.
    induction m as [| l d c r h]; intros; simpl in |- *; auto with emap.
    inversions_all.
    compare_cases key c.
     phys_eq_cases; auto with emap.
     eauto with emap.
     phys_eq_cases; auto with emap.
  Qed.

  Hint Resolve emap_remove_wo : emap.

  Lemma emap_remove_eqv_emap_remove2 :
    forall key : domain,
    forall m : emap_set,
    (emap_wo m) ->
    forall lookup : domain,
    (emap_find lookup (emap_remove key m)) =
    (emap_find lookup (emap_remove2 key m)).
  Proof.
    induction m; intros; inversions_all; auto with emap.
    simpl in |- *.
    compare_cases key d; phys_eq_cases; auto with emap.
     rewrite bal_correct; auto with emap.
       simpl in |- *.
       rewrite <- IHm1; trivial.
       rewrite e.
       trivial.
     rewrite bal_correct; auto with emap.
       rewrite bal_correct; auto with emap.
       simpl in |- *.
       rewrite IHm1; trivial.
     rewrite bal_correct; auto with emap.
       simpl in |- *.
       rewrite <- IHm2.
      rewrite e.
        trivial.
      assumption.
     rewrite bal_correct; auto with emap.
       rewrite bal_correct; auto with emap.
       simpl in |- *.
       rewrite <- IHm2; trivial.
  Qed.

  Lemma emap_remove_correct :
    forall key : domain,
    forall m : emap_set,
    (emap_wo m) ->
    forall lookup : domain,
    (func_remove key (fun lookup1 => emap_find lookup1 m) lookup) =
    (emap_find lookup (emap_remove key m)).
  Proof.
    intros.
    rewrite emap_remove_eqv_emap_remove2; trivial.
    apply emap_remove2_correct; trivial.
  Qed.

  Definition remove (key : domain) (m : t) : t.
  Proof.
    intros.
    destruct m as (m, Hm).
    exists (emap_remove key m).
    apply emap_remove_wo.
    assumption.
  Defined.

  Theorem remove_correct:
    forall key : domain,
    forall m : t,
    forall lookup : domain,
    (func_remove key (fun lookup1 => find lookup1 m) lookup) =
    (find lookup (remove key m)).
  Proof.
    intros.
    unfold find, remove in |- *.
    destruct m as (m, Hm); simpl in |- *.
    apply emap_remove_correct.
    assumption.
  Qed.

  Section Key_cut.
    Variable cutting_key : domain.

    Inductive cut_record : Set :=
      Cut_record_node : emap_set -> (option codomain) ->emap_set ->cut_record.

    Inductive cut_record_wo : cut_record -> Prop :=
      Cut_record_node_wo :
        forall l : emap_set,
        forall c_opt : (option codomain),
        forall r : emap_set,
        (emap_wo l) ->
        (emap_bounded_up cutting_key l) ->
        (emap_bounded_down cutting_key r) ->
        (emap_wo r) ->
        (cut_record_wo (Cut_record_node l c_opt r)).

    Hint Resolve Cut_record_node_wo : emap.

    Inductive cut_record_bounded_up (bound : domain) : cut_record -> Prop :=
      Cut_record_node_bounded_up :
        forall l : emap_set,
        forall c_opt : (option codomain),
        forall r : emap_set,
        (emap_bounded_up bound l) ->
        (emap_bounded_up bound r) ->
        (cut_record_bounded_up bound (Cut_record_node l c_opt r)).

    Inductive cut_record_bounded_down (bound : domain) : cut_record -> Prop :=
      Cut_record_node_bounded_down :
        forall l : emap_set,
        forall c_opt : (option codomain),
        forall r : emap_set,
        (emap_bounded_down bound l) ->
        (emap_bounded_down bound r) ->
        (cut_record_bounded_down bound (Cut_record_node l c_opt r)).

    Hint Resolve Cut_record_node_bounded_up
                 Cut_record_node_bounded_down : emap.

    Fixpoint cut (m : emap_set) : cut_record :=
    match m with
    | Empty_emap => (Cut_record_node Empty_emap None Empty_emap)
    | (Node_emap l1 d1 c1 r1 h1) =>
        match (compare cutting_key d1) with
        | LESS _ _ =>
            let (l2, c2opt, r2) := cut l1 in
            (Cut_record_node l2 c2opt (create r2 d1 c1 r1))
        | EQUAL _ _ =>
            (Cut_record_node l1 (Some c1) r1)
        | GREATER _ _ =>
            let (l2, c2opt, r2) := cut r1 in
            (Cut_record_node (create l1 d1 c1 l2) c2opt r2)
        end
    end.

    Lemma cut_bounded_up :
      forall bound : domain,
      forall m : emap_set,
      (emap_bounded_up bound m) ->
      (cut_record_bounded_up bound (cut m)).
    Proof.
      induction m; simpl in |- *.
       auto with emap.
       intro.
         inversions_all.
         compare_cases cutting_key d; auto with emap.
        assert (Hbound : cut_record_bounded_up bound (cut m1)).
         auto.
         destruct (cut m1).
           inversion_clear Hbound.
           auto with emap.
        assert (Hbound : cut_record_bounded_up bound (cut m2)).
         auto.
         destruct (cut m2).
           inversion_clear Hbound.
           auto with emap.
    Qed.

    Lemma cut_bounded_down :
      forall bound : domain,
      forall m : emap_set,
      (emap_bounded_down bound m) ->
      (cut_record_bounded_down bound (cut m)).
    Proof.
      induction m; simpl in |- *.
       auto with emap.
       intro.
         inversions_all.
         compare_cases cutting_key d; auto with emap.
        assert (Hbound : cut_record_bounded_down bound (cut m1)).
         auto.
         destruct (cut m1).
           inversion_clear Hbound.
           auto with emap.
        assert (Hbound : cut_record_bounded_down bound (cut m2)).
         auto.
         destruct (cut m2).
           inversion_clear Hbound.
           auto with emap.
    Qed.

    Hint Resolve cut_bounded_up cut_bounded_down : emap.

    Lemma cut_wo :
      forall m : emap_set,
      (emap_wo m) ->
      (cut_record_wo (cut m)).
    Proof.
      induction m; simpl in |- *; auto with emap.
      intros; inversions_all.
      compare_cases cutting_key d.
       assert (Hwo : cut_record_wo (cut m1)); auto.
       assert (Hbound : cut_record_bounded_up d (cut m1)); auto with emap.
        destruct (cut m1).
          inversion_clear Hwo.
          inversion_clear Hbound.
          unfold create in |- *.
          eauto with emap.
       eauto with emap.

       eauto with emap.

       assert (Hwo : cut_record_wo (cut m2)); auto.
       assert (Hbound : cut_record_bounded_down d (cut m2)); auto with emap.
        destruct (cut m2).
          inversion_clear Hwo.
          inversion_clear Hbound.
          unfold create in |- *.
          eauto with emap.
    Qed.

    Definition cut_find (lookup : domain) (cr : cut_record) :=
      let (l, c_opt, r) := cr in
      match compare lookup cutting_key with
      | LESS _ _ => (emap_find lookup l)
      | EQUAL _ _ => c_opt
      | GREATER _ _ => (emap_find lookup r)
      end.
      
    Lemma cut_correct :
      forall lookup : domain,
      forall m : emap_set,
      (emap_wo m) ->
      (cut_find lookup (cut m)) = (emap_find lookup m).
    Proof.
      induction m; simpl in |- *; intros.
       compare_cases lookup cutting_key; trivial.
       unfold cut_find in |- *.
         inversions_all.
         assert (Hwo1 : cut_record_wo (cut m1)).
        apply cut_wo; trivial.
        assert (Hwo2 : cut_record_wo (cut m2)).
         apply cut_wo; trivial.
         destruct (cut m1) as (c1l, c1co, c1r).
           destruct (cut m2) as (c2l, c2co, c2r).
           inversion_clear Hwo1.
           inversion_clear Hwo2.
           compare_cases_auto; try rewrite <- IHm1; try rewrite <- IHm2;
            simpl in |- *; order_rewrites; simpl in |- *; 
            auto with emap.
    Qed.

  End Key_cut.

  Section map2_endo.
    Variable f : codomain -> codomain -> codomain.

    Fixpoint emap_map2_neutral (m1 : emap_set) (m2 : emap_set)
      {struct m1} : emap_set :=
      match m1 with
      | Empty_emap => m2
      | (Node_emap m1l d m1c m1r _) =>
        let (m2l, m2co, m2r) := (cut d m2) in
        let l := (emap_map2_neutral m1l m2l) in
        let r := (emap_map2_neutral m1r m2r) in
        let c := match m2co with
                 | Some m2c => (f m1c m2c)
                 | None => m1c
                 end in
        if (sumbool_and (physical_equality l m1l)
           (sumbool_and (physical_equality c m1c)
                        (physical_equality r m1r)))
        then m1
        else (create l d c r)
      end.

    Lemma emap_map2_neutral_bounded_up :
      forall bound : domain,
      forall m1 : emap_set,
      forall m2 : emap_set,
      (emap_bounded_up bound m1) ->
      (emap_bounded_up bound m2) ->
      (emap_bounded_up bound (emap_map2_neutral m1 m2)).
    Proof.
      induction m1; simpl in |- *.
       auto with emap.
       intros; inversions_all.
         assert (Hbound : cut_record_bounded_up bound (cut d m2)).
        apply cut_bounded_up; assumption.
        destruct (cut d m2) as (m2l, m2co, m2r).
          inversion_clear Hbound.
          destruct
           (sumbool_and (physical_equality (emap_map2_neutral m1_1 m2l) m1_1)
              (sumbool_and
                 (physical_equality
                    match m2co with
                    | Some m2c => f c m2c
                    | None => c
                    end c) (physical_equality (emap_map2_neutral m1_2 m2r) m1_2)))
           as [ALL_EQUAL| DIFFERENT].
         auto with emap.
         auto with emap.
    Qed.

    Lemma emap_map2_neutral_bounded_down :
      forall bound : domain,
      forall m1 : emap_set,
      forall m2 : emap_set,
      (emap_bounded_down bound m1) ->
      (emap_bounded_down bound m2) ->
      (emap_bounded_down bound (emap_map2_neutral m1 m2)).
    Proof.
      induction m1; simpl in |- *.
       auto with emap.
       intros; inversions_all.
         assert (Hbound : cut_record_bounded_down bound (cut d m2)).
        apply cut_bounded_down; assumption.
        destruct (cut d m2) as (m2l, m2co, m2r).
          inversion_clear Hbound.
          destruct
           (sumbool_and (physical_equality (emap_map2_neutral m1_1 m2l) m1_1)
              (sumbool_and
                 (physical_equality
                    match m2co with
                    | Some m2c => f c m2c
                    | None => c
                    end c) (physical_equality (emap_map2_neutral m1_2 m2r) m1_2)))
           as [ALL_EQUAL| DIFFERENT].
         auto with emap.
         auto with emap.
    Qed.

    Hint Resolve emap_map2_neutral_bounded_down emap_map2_neutral_bounded_up : emap.

    Lemma emap_map2_neutral_wo :
      forall m1 m2 : emap_set,
      (emap_wo m1) -> (emap_wo m2) -> (emap_wo (emap_map2_neutral m1 m2)).
    Proof.
      induction m1; simpl in |- *.
       auto with emap.
       intros; inversions_all.
         assert (Hwo : cut_record_wo d (cut d m2)).
        apply cut_wo; assumption.
        destruct (cut d m2) as (m2l, m2co, m2r).
          inversion_clear Hwo.
          destruct
           (sumbool_and (physical_equality (emap_map2_neutral m1_1 m2l) m1_1)
              (sumbool_and
                 (physical_equality
                    match m2co with
                    | Some m2c => f c m2c
                    | None => c
                    end c) (physical_equality (emap_map2_neutral m1_2 m2r) m1_2)))
           as [ALL_EQUAL| DIFFERENT].
         auto with emap.
         auto with emap.
    Qed.

    Hint Resolve emap_map2_neutral_wo : emap.

    Definition func_map2_neutral (v1 : domain -> (option codomain))
      (v2 : domain -> (option codomain))
      (lookup : domain) : (option codomain) :=
      match v1 lookup with
      | None => v2 lookup
      | (Some x1) =>
        match v2 lookup with
        | (Some x2) => Some (f x1 x2)
        | None => (Some x1)
        end
      end.

    Lemma emap_map2_neutral_correct :
      forall m1 m2 : emap_set,
      (emap_wo m1) -> (emap_wo m2) ->
      forall lookup : domain,
      (emap_find lookup (emap_map2_neutral m1 m2)) =
      (func_map2_neutral
         (fun lookup1 => emap_find lookup1 m1)
         (fun lookup1 => emap_find lookup1 m2) lookup).
    Proof.
      unfold func_map2_neutral in |- *.
      induction m1; simpl in |- *.
       trivial.
       intros.
         assert (Hwo : cut_record_wo d (cut d m2)).
        apply cut_wo; assumption.
        assert (rew : emap_find lookup m2 = cut_find d lookup (cut d m2)).
         symmetry  in |- *.
           apply cut_correct; assumption.
         destruct (cut d m2) as (m2l, m2d, m2r).
           destruct
            (sumbool_and (physical_equality (emap_map2_neutral m1_1 m2l) m1_1)
               (sumbool_and
                  (physical_equality
                     match m2d with
                     | Some m2c => f c m2c
                     | None => c
                     end c) (physical_equality (emap_map2_neutral m1_2 m2r) m1_2)))
            as [ALL_EQUAL| DIFFERENT].
          destruct ALL_EQUAL as (EQUAL1, EQUAL23).
            destruct EQUAL23 as (EQUAL2, EQUAL3).
            simpl in |- *.
            inversion Hwo.
            inversions_all.
            rewrite rew.
            unfold cut_find in |- *.
            compare_cases lookup d.
           rewrite <- IHm1_1; trivial.
             rewrite EQUAL1.
             trivial.
           destruct m2d.
            rewrite EQUAL2.
              trivial.
            trivial.
           rewrite <- IHm1_2; trivial.
             rewrite EQUAL3.
             trivial.
          clear DIFFERENT.
            rewrite rew.
            unfold cut_find in |- *.
            inversions_all.
            inversion_clear Hwo.
            compare_cases lookup d; order_rewrites.
           rewrite <- IHm1_1; auto with emap.
           destruct m2d; trivial.
           rewrite <- IHm1_2; auto with emap.
    Qed.

    Definition map2_neutral (m1 m2 : t) : t.
    Proof.
      intros.
      destruct m1 as (m1, Hm1wo).
      destruct m2 as (m2, Hm2wo).
      exists (emap_map2_neutral m1 m2).
      auto with emap.
    Defined.

    Theorem map2_neutral_correct :
      forall m1 m2 : t,
      forall lookup : domain,
      (find lookup (map2_neutral m1 m2)) =
      (func_map2_neutral
         (fun lookup1 => find lookup1 m1)
         (fun lookup1 => find lookup1 m2) lookup).
    Proof.
      intros.
      destruct m1 as (m1, Hm1).
      destruct m2 as (m2, Hm2).
      unfold find, map2_neutral in |- *; simpl in |- *.
      apply emap_map2_neutral_correct; assumption.
    Qed.

    Hypothesis f_idem :
      forall x : codomain, (f x x) = x.

    Fixpoint emap_map2_idem (m1 : emap_set) (m2 : emap_set)
      {struct m1} : emap_set :=
      if (physical_equality m1 m2)
      then m1
      else
        match m1 with
        | Empty_emap => m2
        | (Node_emap m1l d m1c m1r _) =>
          let (m2l, m2co, m2r) := (cut d m2) in
          let l := (emap_map2_idem m1l m2l) in
          let r := (emap_map2_idem m1r m2r) in
          let c := match m2co with
                   | Some m2c => (f m1c m2c)
                   | None => m1c
                   end in
          if (sumbool_and (physical_equality l m1l)
             (sumbool_and (physical_equality c m1c)
                          (physical_equality r m1r)))
          then m1
          else (create l d c r)
        end.

    Lemma emap_map2_idem_bounded_up :
      forall bound : domain,
      forall m1 : emap_set,
      forall m2 : emap_set,
      (emap_bounded_up bound m1) ->
      (emap_bounded_up bound m2) ->
      (emap_bounded_up bound (emap_map2_idem m1 m2)).
    Proof.
      induction m1; intros; simpl in |- *.
       destruct (physical_equality Empty_emap m2); auto with emap.
       destruct (physical_equality (Node_emap m1_1 d c m1_2 n) m2).
        trivial.
        inversions_all.
          assert (Hbound : cut_record_bounded_up bound (cut d m2)).
         apply cut_bounded_up; assumption.
         destruct (cut d m2) as (m2l, m2co, m2r).
           inversion_clear Hbound.
           destruct
            (sumbool_and (physical_equality (emap_map2_idem m1_1 m2l) m1_1)
               (sumbool_and
                  (physical_equality
                     match m2co with
                     | Some m2c => f c m2c
                     | None => c
                     end c) (physical_equality (emap_map2_idem m1_2 m2r) m1_2)))
            as [ALL_EQUAL| DIFFERENT].
          auto with emap.
          auto with emap.
    Qed.

    Lemma emap_map2_idem_bounded_down :
      forall bound : domain,
      forall m1 : emap_set,
      forall m2 : emap_set,
      (emap_bounded_down bound m1) ->
      (emap_bounded_down bound m2) ->
      (emap_bounded_down bound (emap_map2_idem m1 m2)).
    Proof.
      induction m1; intros; simpl in |- *.
       destruct (physical_equality Empty_emap m2); auto with emap.
       destruct (physical_equality (Node_emap m1_1 d c m1_2 n) m2).
        trivial.
        inversions_all.
          assert (Hbound : cut_record_bounded_down bound (cut d m2)).
         apply cut_bounded_down; assumption.
         destruct (cut d m2) as (m2l, m2co, m2r).
           inversion_clear Hbound.
           destruct
            (sumbool_and (physical_equality (emap_map2_idem m1_1 m2l) m1_1)
               (sumbool_and
                  (physical_equality
                     match m2co with
                     | Some m2c => f c m2c
                     | None => c
                     end c) (physical_equality (emap_map2_idem m1_2 m2r) m1_2)))
            as [ALL_EQUAL| DIFFERENT].
          auto with emap.
          auto with emap.
    Qed.

    Hint Resolve emap_map2_idem_bounded_up emap_map2_idem_bounded_down: emap.

    Lemma emap_map2_idem_wo :
      forall m1 : emap_set,
      forall m2 : emap_set,
      (emap_wo m1) ->
      (emap_wo m2) ->
      (emap_wo (emap_map2_idem m1 m2)).
    Proof.
      induction m1; intros; simpl in |- *.
       destruct (physical_equality Empty_emap m2); auto with emap.
       destruct (physical_equality (Node_emap m1_1 d c m1_2 n) m2).
        trivial.
        inversions_all.
          assert (Hwo : cut_record_wo d (cut d m2)).
         apply cut_wo; assumption.
         destruct (cut d m2) as (m2l, m2co, m2r).
           inversion_clear Hwo.
           destruct
            (sumbool_and (physical_equality (emap_map2_idem m1_1 m2l) m1_1)
               (sumbool_and
                  (physical_equality
                     match m2co with
                     | Some m2c => f c m2c
                     | None => c
                     end c) (physical_equality (emap_map2_idem m1_2 m2r) m1_2)))
            as [ALL_EQUAL| DIFFERENT].
          auto with emap.
          auto with emap.
    Qed.

    Hint Resolve emap_map2_idem_wo : emap.

    Lemma emap_map2_idem_correct :
      forall m1 m2 : emap_set,
      (emap_wo m1) -> (emap_wo m2) ->
      forall lookup : domain,
      (emap_find lookup (emap_map2_idem m1 m2)) =
      (func_map2_neutral
         (fun lookup1 => emap_find lookup1 m1)
         (fun lookup1 => emap_find lookup1 m2) lookup).
    Proof.
      unfold func_map2_neutral.
      induction m1; simpl in |- *.
       intros.
         destruct (physical_equality Empty_emap m2).
        rewrite e.
          trivial.
        trivial.
       intros.
         assert (Hwo : cut_record_wo d (cut d m2)).
        apply cut_wo; assumption.
        assert (rew : emap_find lookup m2 = cut_find d lookup (cut d m2)).
         symmetry  in |- *.
           apply cut_correct; assumption.
         destruct (cut d m2) as (m2l, m2d, m2r).
           destruct (physical_equality (Node_emap m1_1 d c m1_2 n) m2)
      as [PHYS_EQUAL| DIFFERENT].
          simpl in |- *.
      replace (emap_find lookup m2) with
       (emap_find lookup (Node_emap m1_1 d c m1_2 n)).
           simpl in |- *.
       compare_cases lookup d.
      destruct (emap_find lookup m1_1).
       rewrite f_idem; trivial.
       trivial.
      rewrite f_idem; trivial.
      destruct (emap_find lookup m1_2).
       rewrite f_idem; trivial.
       trivial.
           rewrite PHYS_EQUAL; trivial.
          clear DIFFERENT.
      destruct
       (sumbool_and (physical_equality (emap_map2_idem m1_1 m2l) m1_1)
          (sumbool_and
             (physical_equality
                match m2d with
                | Some m2c => f c m2c
                | None => c
                end c) (physical_equality (emap_map2_idem m1_2 m2r) m1_2)))
       as [ALL_EQUAL| DIFFERENT].
           destruct ALL_EQUAL as (EQUAL1, EQUAL23).
       destruct EQUAL23 as (EQUAL2, EQUAL3).
       simpl in |- *.
       inversion Hwo.
       inversions_all.
       rewrite rew.
       unfold cut_find in |- *.
       compare_cases lookup d.
      rewrite <- IHm1_1; trivial.
        rewrite EQUAL1.
        trivial.
      destruct m2d.
       rewrite EQUAL2.
         trivial.
       trivial.
      rewrite <- IHm1_2; trivial.
        rewrite EQUAL3.
        trivial.
           clear DIFFERENT.
       rewrite rew.
       unfold cut_find in |- *.
       inversions_all.
       inversion_clear Hwo.
       compare_cases lookup d; order_rewrites.
      rewrite <- IHm1_1; auto with emap.
      destruct m2d; trivial.
      rewrite <- IHm1_2; auto with emap.
    Qed.

    Definition map2_idem (m1 m2 : t) : t.
    Proof.
      intros.
      destruct m1 as (m1, Hm1wo).
      destruct m2 as (m2, Hm2wo).
      exists (emap_map2_idem m1 m2).
      auto with emap.
    Defined.

    Theorem map2_idem_correct :
      forall m1 m2 : t,
      forall lookup : domain,
      (find lookup (map2_idem m1 m2)) =
      (func_map2_neutral
         (fun lookup1 => find lookup1 m1)
         (fun lookup1 => find lookup1 m2) lookup).
    Proof.
      intros.
      destruct m1 as (m1, Hm1).
      destruct m2 as (m2, Hm2).
      unfold find, map2_idem in |- *; simpl in |- *.
      apply emap_map2_idem_correct; assumption.
    Qed.
  End map2_endo.

  Section Emap_foldi_right.
    Variable beta : Set.
    Variable f : domain -> codomain -> beta -> beta.

    Fixpoint emap_foldi_right (m : emap_set) (init : beta) {struct m} : beta :=
      match m with
      | Empty_emap => init
      | (Node_emap l d c r _) =>
          (emap_foldi_right l (f d c (emap_foldi_right r init)))
      end.

   Fixpoint prepend_list (lst1 : (alist.alist_set codomain))
     (lst2 : (alist.alist_set codomain))
     { struct lst1 } : (alist.alist_set codomain) :=
     match lst1 with
     | alist.Empty_alist => lst2
     | (alist.Cons_alist c1 d1 tail1) =>
       (alist.Cons_alist codomain c1 d1 (prepend_list tail1 lst2))
     end.

    Lemma alist_foldi_right_prepend:
      forall lst1 : (alist.alist_set codomain),
      forall lst2 : (alist.alist_set codomain),
      forall init : beta,
      alist.alist_foldi_right codomain beta f lst1
       (alist.alist_foldi_right codomain beta f lst2 init) =
      alist.alist_foldi_right codomain beta f (prepend_list lst1 lst2) init.
    Proof.
      induction lst1; simpl in |- *.
       trivial.
       intros.
         rewrite IHlst1.
         trivial.
    Qed.

     Lemma alist_prepend_emap_to_assoc_list_swap :
       forall m : emap_set,
       forall lst1 lst2 : (alist.alist_set codomain),
       (prepend_list (prepend_emap_to_assoc_list m lst1) lst2) =
       (prepend_emap_to_assoc_list m (prepend_list lst1 lst2)).
     Proof.
      induction m; simpl in |- *; intros.
       trivial.
       rewrite IHm1.
         rewrite <- IHm2.
         simpl in |- *.
         trivial.
    Qed.

   Lemma emap_foldi_right_prepend_emap_to_assoc_list :
     forall m : emap_set,
     forall lst1 lst2 : (alist.alist_set codomain),
     forall init : beta,
     (alist.alist_foldi_right codomain beta f
       (prepend_emap_to_assoc_list m lst1)
       (alist.alist_foldi_right codomain beta f lst2 init)) =
     (alist.alist_foldi_right codomain beta f
       (prepend_emap_to_assoc_list m
          (prepend_list lst1 lst2)) init).
   Proof.
      induction m; intros; simpl in |- *.
       apply alist_foldi_right_prepend.
       rewrite IHm1.
         simpl in |- *.
         rewrite alist_prepend_emap_to_assoc_list_swap.
         trivial.
   Qed.

   Lemma emap_foldi_right_correct :
     forall m : emap_set,
     forall init : beta,
     (emap_foldi_right m init) =
     (alist.alist_foldi_right codomain beta f (emap_to_assoc_list m) init).
   Proof.
      induction m as [| l Hl d c r Hr h]; simpl in |- *; intros.
       trivial.
       unfold emap_to_assoc_list in |- *.
         simpl in |- *.
         replace (prepend_emap_to_assoc_list r (alist.Empty_alist codomain)) with
          (emap_to_assoc_list r).
        rewrite Hl.
          rewrite Hr.
          replace
           (alist.alist_foldi_right codomain beta f (emap_to_assoc_list l)
        (f d c
           (alist.alist_foldi_right codomain beta f 
              (emap_to_assoc_list r) init))) with
           (alist.alist_foldi_right codomain beta f (emap_to_assoc_list l)
        (alist.alist_foldi_right codomain beta f
           (alist.Cons_alist codomain d c (emap_to_assoc_list r)) init)).
         unfold emap_to_assoc_list in |- *.
           rewrite emap_foldi_right_prepend_emap_to_assoc_list.
           simpl in |- *.
           trivial.
         simpl in |- *.
           trivial.
        unfold emap_to_assoc_list in |- *.
          trivial.
   Qed.

   Definition foldi_right (m : t) (init : beta) :=
     emap_foldi_right (proj1_sig m) init.

   Theorem foldi_right_correct1 :
     forall m : t,
     forall init : beta,
     (foldi_right m init) =
     (alist.foldi_right codomain beta f (to_assoc_list m) init).
   Proof.
    intro m.
    destruct m as (m, Hm).
    unfold to_assoc_list, foldi_right, alist.foldi_right in |- *; simpl in |- *.
    apply emap_foldi_right_correct.
   Qed.

   Theorem foldi_right_correct :
     forall m : t,
     exists l : (alist.t codomain),
     (forall lookup : domain,
       (find lookup m) = (alist.find codomain lookup l)) /\
     forall init : beta,
     (foldi_right m init) =
     (alist.foldi_right codomain beta f l init).
    Proof.
      intro m.
      exists (to_assoc_list m).
      split.
       apply to_assoc_list_correct.
       apply foldi_right_correct1.
    Qed.
  End Emap_foldi_right.

  Section Comparison.
     Variable  codomain_le : codomain -> codomain -> Prop.
     Hypothesis codomain_le_refl :
       forall x : codomain, (codomain_le x x).

     Definition emap_le (l1 l2 : emap_set) := forall x : domain,
       match emap_find x l2 with
       | None => True
       | Some v2 =>
         match emap_find x l1 with
         | None => False
         | Some v1 => codomain_le v1 v2
         end
       end.

     Lemma emap_le_refl : forall l : emap_set, (emap_le l l).
      unfold emap_le in |- *.
      intros.
      destruct (emap_find x l).
       apply codomain_le_refl.
       trivial.
     Qed.

     Variable codomain_le_decide :
       forall v1 v2 : codomain,
       { codomain_le v1 v2 } + { ~(codomain_le v1 v2) }.

     Definition emap_le_decide:
       forall l1 l2 : emap_set,
       (emap_wo l1) -> (emap_wo l2) ->
       { emap_le l1 l2 } + { ~(emap_le l1 l2) }.

     Proof.
      intros l1 l2.
         destruct (physical_equality l1 l2) as [e| dummy].
        left.
          rewrite e.
          apply emap_le_refl.
        clear dummy.
          generalize l2.
          clear l2.
          generalize l1.
          clear l1.
      induction l1.
       destruct l2.
        left.
          unfold emap_le in |- *.
          simpl in |- *.
          trivial.
        intros Hl1wo Hl2wo.
          right.
          inversion_clear Hl2wo.
          unfold emap_le in |- *.
          intro Habsurd.
          generalize (Habsurd d).
          simpl in |- *.
          rewrite le_refl_EQUAL.
          trivial.
       intros l2 Hl1wo Hl2wo.
         generalize (fun lookup => cut_correct d lookup l2 Hl2wo).
         generalize (cut_wo d l2 Hl2wo).
         destruct (cut d l2) as (l2l, l2copt, l2r).
         intros Hcutwo Hcutcorrect.
         assert (Hlcompare : {emap_le l1_1 l2l} + {~ emap_le l1_1 l2l}).
        apply IHl1_1.
         inversion_clear Hl1wo; assumption.
         inversion_clear Hcutwo; assumption.
        assert (Hrcompare : {emap_le l1_2 l2r} + {~ emap_le l1_2 l2r}).
         apply IHl1_2.
          inversion_clear Hl1wo; assumption.
          inversion_clear Hcutwo; assumption.
         destruct l2copt as [l2c| ].
          destruct
           (sumbool_and (codomain_le_decide c l2c)
              (sumbool_and Hlcompare Hrcompare)) as [YES| NO].
           left.
             unfold emap_le in |- *.
             unfold emap_le in YES.
             intro.
             rewrite <- Hcutcorrect.
             clear Hcutcorrect.
             simpl in |- *.
             destruct YES as (YESc, (YESl, YESr)).
             compare_cases x d.
            apply YESl.
            apply YESr.
           right.
             destruct NO as [NO| [NO| NO]]; intro Habsurd; apply NO;
              unfold emap_le in Habsurd.
            generalize (Habsurd d).
              rewrite <- Hcutcorrect.
              simpl in |- *.
              rewrite le_refl_EQUAL.
              rewrite le_refl_EQUAL.
              trivial.
            unfold emap_le in |- *.
              intro x.
              generalize (Habsurd x).
              rewrite <- Hcutcorrect.
              simpl in |- *.
              compare_cases x d.
             rewrite emap_find_bounded_up.
              trivial.
              inversion_clear Hcutwo; trivial.
                eauto with emap.
             replace (emap_find x l2l) with (None (A:=codomain)).
              trivial.
              symmetry  in |- *.
                apply emap_find_bounded_up.
                apply emap_bounded_up_trans with (bound1 := d).
               assumption.
               inversion_clear Hcutwo; assumption.
            unfold emap_le in |- *.
              intro x.
              generalize (Habsurd x).
              rewrite <- Hcutcorrect.
              simpl in |- *.
              compare_cases x d.
             replace (emap_find x l2r) with (None (A:=codomain)).
              trivial.
              symmetry  in |- *.
                apply emap_find_bounded_down.
                apply emap_bounded_down_trans with (bound2 := d).
               assumption.
               inversion_clear Hcutwo.
                 assumption.
             rewrite emap_find_bounded_down.
              trivial.
              inversion_clear Hcutwo.
                apply emap_bounded_down_trans with (bound2 := d); assumption.
          destruct (sumbool_and Hlcompare Hrcompare) as [YES| NO].
           destruct YES as (YESl, YESr).
             left.
             unfold emap_le in |- *.
             unfold emap_le in YESl.
             unfold emap_le in YESr.
             intro.
             rewrite <- Hcutcorrect.
             simpl in |- *.
             compare_cases x d.
            apply YESl.
            apply YESr.
           right.
             destruct NO as [NO| NO].
            unfold emap_le in |- *.
              unfold emap_le in NO.
              intro Habsurd.
              apply NO.
              intro.
              generalize (Habsurd x).
              rewrite <- Hcutcorrect.
              simpl in |- *.
              compare_cases x d.
             rewrite emap_find_bounded_up.
              trivial.
              apply emap_bounded_up_trans with (bound1 := d).
               assumption.
               inversion_clear Hcutwo; assumption.
             replace (emap_find x l2l) with (None (A:=codomain)).
              trivial.
              symmetry  in |- *.
                apply emap_find_bounded_up.
                apply emap_bounded_up_trans with (bound1 := d).
               assumption.
               inversion_clear Hcutwo; assumption.
            unfold emap_le in |- *.
              unfold emap_le in NO.
              intro Habsurd.
              apply NO.
              intro.
              generalize (Habsurd x).
              rewrite <- Hcutcorrect.
              simpl in |- *.
              compare_cases x d.
             replace (emap_find x l2r) with (None (A:=codomain)).
              trivial.
              symmetry  in |- *.
                apply emap_find_bounded_down.
                apply emap_bounded_down_trans with (bound2 := d).
               assumption.
               inversion_clear Hcutwo; assumption.
             rewrite emap_find_bounded_down.
              trivial.
              apply emap_bounded_down_trans with (bound2 := d).
               assumption.
               inversion_clear Hcutwo; assumption.
    Defined.

    Definition pointwise_le (l1 l2 : t) := forall x : domain,
       match find x l2 with
       | None => True
       | Some v2 =>
         match find x l1 with
         | None => False
         | Some v1 => codomain_le v1 v2
         end
       end.

    Definition pointwise_le_decide:
       forall l1 l2 : t,
       { pointwise_le l1 l2 } + { ~(pointwise_le l1 l2) }.
    Proof.
      destruct l1 as (l1, l1wo).
      destruct l2 as (l2, l2wo).
      unfold pointwise_le, find in |- *.
      simpl in |- *.
      exact (emap_le_decide l1 l2 l1wo l2wo).
    Defined.
  End Comparison.
End Emap_set.

Hint Resolve
    Empty_emap_bounded_down
    Node_emap_bounded_down
    Empty_emap_bounded_up
    Node_emap_bounded_up
    Empty_emap_wo
    Node_emap_wo
    emap_bounded_up_trans
    emap_bounded_down_trans : emap.

Hint Resolve create_bounded_up create_bounded_down create_wo : emap.
Hint Resolve merge_bounded_up merge_bounded_down merge_wo : emap.
Hint Resolve cut_bounded_up cut_bounded_down cut_wo : emap.
Hint Resolve merge_correct1 merge_correct2 merge_correctM : emap.

Ltac inversions_all := repeat
  match goal with
  | H : emap_bounded_up _ _ (Node_emap _ _ _ _ _ _) |- _ => inversion_clear H
  | H : emap_bounded_down _ _ (Node_emap _ _ _ _ _ _) |- _ => inversion_clear H
  | H : emap_wo _ (Node_emap _ _ _ _ _ _) |- _ => inversion_clear H
  | H : cut_record_bounded_up _ _ _ |- _ => inversion_clear H
  | H : cut_record_bounded_down _ _ _ |- _ => inversion_clear H
  | H : cut_record_wo _ _ _ |- _ => inversion_clear H
  end.



Section widening.
  Variable codomain : Set.
  Variable  codomain_le : codomain -> codomain -> Prop.
  Hypothesis codomain_le_refl :
       forall x : codomain, (codomain_le x x).
  Variable codomain_widening_chain : (widening_chains.widening_chain codomain codomain_le).

  Inductive comp_widening_chain (S : Set) : Set :=
    comp_widening_chain_intro : forall x : S,
      (forall y : S,
        (option (comp_widening_chain S))) -> (comp_widening_chain S).



  Fixpoint add_phys_eq_test_to_widening
    (old_widening : (comp_widening_chain (emap_set codomain))) :
      (comp_widening_chain (emap_set codomain)) :=
    let (point, old_step) := old_widening in
    (comp_widening_chain_intro (emap_set codomain) point
      (fun y =>
        if (physical_equality point y)
        then None
        else
          match old_step y with
          | None => None
          | Some next => Some (add_phys_eq_test_to_widening next)
          end)).

  Fixpoint at_all_steps_of_chain (X : Set)
    (Ppoint : X->Prop) (Py : X->Prop)
    (chain : (comp_widening_chain X)) {struct chain} : Prop :=
    let (point, step) := chain in
    (Ppoint point) /\
    (forall y : X, (Py y) ->
     match step y with
     | None => True
     | Some next => (at_all_steps_of_chain X Ppoint Py next)
     end).

Definition option_fold (A:Set) (B:Type) (f:A->B) (b:B) o :=
  match o with
  | None => b
  | Some x => f x
  end.

Implicit Arguments option_fold [A B].

Definition comp_widening_chain_rect'
  (S:Set)
  (P : comp_widening_chain S -> Type)
  (f : forall (x:S) (c: S -> option (comp_widening_chain S)),
       (forall y:S, option_fold P True (c y)) ->
       P (comp_widening_chain_intro S x c))
  :=
  fix F (c : comp_widening_chain S) : P c :=
    match c return (P c) with
    | comp_widening_chain_intro x d =>
      f x d (fun y => 
        match d y as z return option_fold P True z with
        | None => I
        | Some x => F x
        end)
    end.

  Lemma add_phys_eq_test_to_widening_bounded_up:
    forall bound : domain,
    forall old : (comp_widening_chain (emap_set codomain)),
    (at_all_steps_of_chain (emap_set codomain)
      (emap_bounded_up codomain bound)
      (emap_bounded_up codomain bound) old) ->
    (at_all_steps_of_chain (emap_set codomain)
      (emap_bounded_up codomain bound)
      (emap_bounded_up codomain bound)
      (add_phys_eq_test_to_widening old)).
  Proof.
    intro.
    induction old as (point, step, Hnext) using comp_widening_chain_rect'.
    unfold option_fold in *.
    simpl in |- *.
    intro Hbound.
    destruct Hbound as (Hbound1, Hbound2).
    split.
     assumption.
     intros y Hy.
       generalize (Hbound2 y Hy).
       clear Hbound2.
       intro Hbound2.
       destruct (physical_equality point y).
      trivial.
      generalize (Hnext y).
        destruct (step y) as [next_y| ]; trivial.
        intro inductive_predicate.
        apply inductive_predicate.
        exact Hbound2.
  Qed.

(*
  Proof.
    intro.
    refine
     (fix inductive_predicate (old : comp_widening_chain (emap_set codomain)) :
        at_all_steps_of_chain (emap_set codomain)
          (emap_bounded_up codomain bound) (emap_bounded_up codomain bound) old ->
        at_all_steps_of_chain (emap_set codomain)
          (emap_bounded_up codomain bound) (emap_bounded_up codomain bound)
          (add_phys_eq_test_to_widening old) := _).
    destruct old as (point, old_step).
    simpl in |- *.
    intro Hbound.
    destruct Hbound as (Hbound1, Hbound2).
    split.
     assumption.
     intros y Hy.
       generalize (Hbound2 y Hy).
       clear Hbound2.
       intro Hbound2.
       destruct (physical_equality point y).
      trivial.
      destruct (old_step y) as [next_y| ]; trivial.
        apply inductive_predicate.
        exact Hbound2.
  Qed.
*)

  Definition split_widening (split_point : domain) :=
    fix split_widening_l
      (left_widening : (comp_widening_chain (emap_set codomain))) :
      (comp_widening_chain codomain) ->
      (comp_widening_chain (emap_set codomain)) ->
      (comp_widening_chain (emap_set codomain)) :=
    fix split_widening_c
      (center_widening:(comp_widening_chain codomain)):
      (comp_widening_chain (emap_set codomain)) ->
      (comp_widening_chain (emap_set codomain)) :=
    fix split_widening_r
      (right_widening : (comp_widening_chain (emap_set codomain))) :
      (comp_widening_chain (emap_set codomain)) :=
      let (point_l, step_l) := left_widening in
      let (point_c, step_c) := center_widening in
      let (point_r, step_r) := right_widening in
      (comp_widening_chain_intro (emap_set codomain)
        (create codomain point_l split_point point_c point_r)
        (fun (y : (emap_set codomain)) =>
          let (y_l, y_c_opt, y_r) := (cut codomain split_point y) in
          match y_c_opt with
          | None => None
          | (Some y_c) =>
            match (step_l y_l) with
	    | Some next_l =>
              Some (split_widening_l next_l
                (match step_c y_c with
                | Some next_c => next_c
                | None => center_widening
                end)
		(match step_r y_r with
		| Some next_r => next_r
		| None => right_widening
                end))
	    | None =>
              	match (step_c y_c) with
                | Some next_c =>
                  Some (split_widening_c next_c
		    (match step_r y_r with
		    | Some next_r => next_r
		    | None => right_widening
                    end))
                | None =>
		  match (step_r y_r) with
                  | Some next_r =>
		    Some (split_widening_r next_r)
                  | None => None
                  end
                end
            end
          end)).

  Lemma split_widening_bounded_up:
    forall bound : domain,
    forall split_point : domain,
    (domain_le split_point bound) ->
    ~ (domain_le bound split_point) ->
    forall left_widening : (comp_widening_chain (emap_set codomain)),
    forall center_widening : (comp_widening_chain codomain),
    forall right_widening : (comp_widening_chain (emap_set codomain)),
    (at_all_steps_of_chain (emap_set codomain)
      (emap_bounded_up codomain bound)
      (emap_bounded_up codomain bound) left_widening) ->
    (at_all_steps_of_chain (emap_set codomain)
      (emap_bounded_up codomain bound)
      (emap_bounded_up codomain bound) right_widening) ->
    (at_all_steps_of_chain (emap_set codomain)
      (emap_bounded_up codomain bound)
      (emap_bounded_up codomain bound)
      (split_widening split_point
         left_widening center_widening right_widening)).
  Proof.
    intros bound split_point Hsplit_point1 Hsplit_point2.
    induction left_widening as (point_l, step_l, Hrec_l) using
     comp_widening_chain_rect'.
    induction center_widening as (point_c, step_c, Hrec_c) using
     comp_widening_chain_rect'.
    induction right_widening as (point_r, step_r, Hrec_r) using
     comp_widening_chain_rect'.
    intros Hbound_l Hbound_r.
    case Hbound_l; intros Hbound_l1 Hbound_l2.
    case Hbound_r; intros Hbound_r1 Hbound_r2.
    simpl in |- *.
    split.
     auto with emap.
     intros y Hy.
       generalize (cut_bounded_up codomain split_point bound y Hy).
       destruct (cut codomain split_point y) as (y_l, y_c_opt, y_r).
       clear y Hy.
       intro Hy.
       inversion_clear Hy.
       destruct y_c_opt as [y_c| ]; trivial.
       unfold option_fold in *.
       generalize (Hrec_l y_l); clear Hrec_l; intro Hrec_l.
       generalize (Hrec_c y_c); clear Hrec_c; intro Hrec_c.
       generalize (Hrec_r y_r); clear Hrec_r; intro Hrec_r.
       generalize (Hbound_l2 y_l H); clear Hbound_l2; intro Hbound_l2.
       generalize (Hbound_r2 y_r H0); clear Hbound_r2; intro Hbound_r2.
       destruct (step_l y_l) as [next_l| ].
      apply Hrec_l.
       exact Hbound_l2.
       destruct (step_r y_r) as [next_r| ].
        exact Hbound_r2.
        assumption.
      destruct (step_c y_c) as [next_c| ].
       apply Hrec_c.
        assumption.
        destruct (step_r y_r) as [next_r| ].
         assumption.
         assumption.
       destruct (step_r y_r) as [next_r| ]; trivial.
         apply Hrec_r.
        assumption.
        assumption.
  Qed.

  Lemma split_widening_bounded_down:
    forall bound : domain,
    forall split_point : domain,
    (domain_le bound split_point) ->
    ~ (domain_le split_point bound) ->
    forall left_widening : (comp_widening_chain (emap_set codomain)),
    forall center_widening : (comp_widening_chain codomain),
    forall right_widening : (comp_widening_chain (emap_set codomain)),
    (at_all_steps_of_chain (emap_set codomain)
      (emap_bounded_down codomain bound)
      (emap_bounded_down codomain bound) left_widening) ->
    (at_all_steps_of_chain (emap_set codomain)
      (emap_bounded_down codomain bound)
      (emap_bounded_down codomain bound) right_widening) ->
    (at_all_steps_of_chain (emap_set codomain)
      (emap_bounded_down codomain bound)
      (emap_bounded_down codomain bound)
      (split_widening split_point
         left_widening center_widening right_widening)).
  Proof.
    intros bound split_point Hsplit_point1 Hsplit_point2.
    induction left_widening as (point_l, step_l, Hrec_l) using
     comp_widening_chain_rect'.
    induction center_widening as (point_c, step_c, Hrec_c) using
     comp_widening_chain_rect'.
    induction right_widening as (point_r, step_r, Hrec_r) using
     comp_widening_chain_rect'.
    intros Hbound_l Hbound_r.
    case Hbound_l; intros Hbound_l1 Hbound_l2.
    case Hbound_r; intros Hbound_r1 Hbound_r2.
    simpl in |- *.
    split.
     auto with emap.
     intros y Hy.
       generalize (cut_bounded_down codomain split_point bound y Hy).
       destruct (cut codomain split_point y) as (y_l, y_c_opt, y_r).
       clear y Hy.
       intro Hy.
       inversion_clear Hy.
       destruct y_c_opt as [y_c| ]; trivial.
       unfold option_fold in *.
       generalize (Hrec_l y_l); clear Hrec_l; intro Hrec_l.
       generalize (Hrec_c y_c); clear Hrec_c; intro Hrec_c.
       generalize (Hrec_r y_r); clear Hrec_r; intro Hrec_r.
       generalize (Hbound_l2 y_l H); clear Hbound_l2; intro Hbound_l2.
       generalize (Hbound_r2 y_r H0); clear Hbound_r2; intro Hbound_r2.
       destruct (step_l y_l) as [next_l| ].
      apply Hrec_l.
       exact Hbound_l2.
       destruct (step_r y_r) as [next_r| ].
        exact Hbound_r2.
        assumption.
      destruct (step_c y_c) as [next_c| ].
       apply Hrec_c.
        assumption.
        destruct (step_r y_r) as [next_r| ].
         assumption.
         assumption.
       destruct (step_r y_r) as [next_r| ]; trivial.
         apply Hrec_r.
        assumption.
        assumption.
  Qed.

  Lemma split_widening_wo:
    forall split_point : domain,
    forall left_widening : (comp_widening_chain (emap_set codomain)),
    forall center_widening : (comp_widening_chain codomain),
    forall right_widening : (comp_widening_chain (emap_set codomain)),
    (at_all_steps_of_chain (emap_set codomain)
      (emap_wo codomain) (emap_wo codomain) left_widening) ->
    (at_all_steps_of_chain (emap_set codomain)
      (emap_bounded_up codomain split_point)
      (emap_bounded_up codomain split_point) left_widening) ->
    (at_all_steps_of_chain (emap_set codomain)
      (emap_bounded_down codomain split_point)
      (emap_bounded_down codomain split_point) right_widening) ->
    (at_all_steps_of_chain (emap_set codomain)
      (emap_wo codomain) (emap_wo codomain) right_widening) ->
    (at_all_steps_of_chain (emap_set codomain)
      (emap_wo codomain) (emap_wo codomain)
      (split_widening split_point
         left_widening center_widening right_widening)).
  Proof.
    intro split_point.
    induction left_widening as (point_l, step_l, Hrec_l) using
     comp_widening_chain_rect'.
    induction center_widening as (point_c, step_c, Hrec_c) using
     comp_widening_chain_rect'.
    induction right_widening as (point_r, step_r, Hrec_r) using
     comp_widening_chain_rect'.
    intros Hwo_l Hbound_l Hbound_r Hwo_r.
    case Hwo_l; intros Hwo_l1 Hwo_l2.
    case Hbound_l; intros Hbound_l1 Hbound_l2.
    case Hbound_r; intros Hbound_r1 Hbound_r2.
    case Hwo_r; intros Hwo_r1 Hwo_r2.
    simpl in |- *.
    split.
     apply create_wo; assumption.
     intros y Hy.
       generalize (cut_wo codomain split_point y Hy).
       destruct (cut codomain split_point y) as (y_l, y_c_opt, y_r).
       clear Hy.
       intro Hy.
       inversion_clear Hy.
       destruct y_c_opt as [y_c| ]; trivial.
       unfold option_fold in *.
       generalize (Hrec_l y_l); clear Hrec_l; intro Hrec_l.
       generalize (Hrec_c y_c); clear Hrec_c; intro Hrec_c.
       generalize (Hrec_r y_r); clear Hrec_r; intro Hrec_r.
       generalize (Hwo_l2 y_l H); clear Hwo_l2; intro Hwo_l2.
       generalize (Hbound_l2 y_l H0); clear Hbound_l2; intro Hbound_l2.
       generalize (Hbound_r2 y_r H1); clear Hbound_r2; intro Hbound_r2.
       generalize (Hwo_r2 y_r H2); clear Hwo_r2; intro Hwo_r2.
       destruct (step_l y_l) as [next_l| ].
      apply Hrec_l; trivial.
       destruct (step_r y_r) as [next_r| ]; assumption.
       destruct (step_r y_r) as [next_r| ]; assumption.
      destruct (step_c y_c) as [next_c| ].
       apply Hrec_c; trivial.
        destruct (step_r y_r) as [next_r| ]; assumption.
        destruct (step_r y_r) as [next_r| ]; assumption.
       destruct (step_r y_r) as [next_r| ]; trivial.
         apply Hrec_r; trivial.
  Qed.
(*
  Fixpoint comp_widening_correct (S : Set) (S_le : S -> S -> Prop)
     (S_filter : S -> Prop)
     (chain : (comp_widening_chain S)) { struct chain } : Prop :=
     let (point, step) := chain in
     forall y : S, (S_filter y) ->
       match step y with
       | Some next => comp_widening_correct S S_le S_filter next
       | None => S_le y point
       end.

  Lemma split_widening_correct:
    forall split_point : domain,
    forall left_widening : (comp_widening_chain (emap_set codomain)),
    forall center_widening : (comp_widening_chain codomain),
    forall right_widening : (comp_widening_chain (emap_set codomain)),
    (at_all_steps_of_chain (emap_set codomain)
      (emap_wo codomain) (emap_wo codomain) left_widening) ->
    (at_all_steps_of_chain (emap_set codomain)
      (emap_bounded_up codomain split_point)
      (emap_bounded_up codomain split_point) left_widening) ->
    (at_all_steps_of_chain (emap_set codomain)
      (emap_bounded_down codomain split_point)
      (emap_bounded_down codomain split_point) right_widening) ->
    (at_all_steps_of_chain (emap_set codomain)
      (emap_wo codomain) (emap_wo codomain) right_widening) ->
    (comp_widening_correct (emap_set codomain) (emap_le codomain codomain_le)
      (emap_wo codomain) left_widening) ->
    (comp_widening_correct codomain codomain_le
      (fun _ => True) center_widening) ->
    (comp_widening_correct (emap_set codomain) (emap_le codomain codomain_le)
      (emap_wo codomain)  right_widening) ->
    (comp_widening_correct (emap_set codomain) (emap_le codomain codomain_le)
      (emap_wo codomain) (split_widening split_point
         left_widening center_widening right_widening)).
  Proof.
intro split_point.
induction left_widening as (point_l, step_l, Hrec_l) using
 comp_widening_chain_rect'.
induction center_widening as (point_c, step_c, Hrec_c) using
 comp_widening_chain_rect'.
induction right_widening as (point_r, step_r, Hrec_r) using
 comp_widening_chain_rect'.
intros Hwo_l Hbound_l Hbound_r Hwo_r Hcorrect_l Hcorrect_c Hcorrect_r.
case Hwo_l; intros Hwo_l1 Hwo_l2.
case Hbound_l; intros Hbound_l1 Hbound_l2.
case Hbound_r; intros Hbound_r1 Hbound_r2.
case Hwo_r; intros Hwo_r1 Hwo_r2.
simpl in |- *.
intros y Hy.
generalize (cut_wo codomain split_point y Hy).
generalize (fun lookup => cut_correct codomain split_point lookup y Hy).
destruct (cut codomain split_point y) as (y_l, y_c_opt, y_r).
clear Hy.
intros Hcut Hywo.
inversion_clear Hywo.
destruct y_c_opt as [y_c| ].
 unfold option_fold in *.
   generalize (Hrec_l y_l); clear Hrec_l; intro Hrec_l.
   generalize (Hrec_c y_c); clear Hrec_c; intro Hrec_c.
   generalize (Hrec_r y_r); clear Hrec_r; intro Hrec_r.
   generalize (Hwo_l2 y_l H); clear Hwo_l2; intro Hwo_l2.
   generalize (Hbound_l2 y_l H0); clear Hbound_l2; intro Hbound_l2.
   generalize (Hbound_r2 y_r H1); clear Hbound_r2; intro Hbound_r2.
   generalize (Hwo_r2 y_r H2); clear Hwo_r2; intro Hwo_r2.
   unfold comp_widening_correct in Hcorrect_l, Hcorrect_c, Hcorrect_r.
   generalize (Hcorrect_l y_l H); intro Hcorrect_l2.
   generalize (Hcorrect_c y_c I); intro Hcorrect_c2.
   generalize (Hcorrect_r y_r H2); intro Hcorrect_r2.
   destruct (step_l y_l) as [next_l| ].
  apply Hrec_l; trivial.
   destruct (step_r y_r) as [next_r| ]; assumption.
   destruct (step_r y_r) as [next_r| ]; assumption.
   destruct (step_c y_c) as [next_c| ]; trivial.
   destruct (step_r y_r) as [next_r| ]; assumption.
  destruct (step_c y_c) as [next_c| ]; trivial.
   apply Hrec_c; destruct (step_r y_r) as [next_r| ]; trivial.
   destruct (step_r y_r) as [next_r| ]; trivial.
    apply Hrec_r; trivial.

    unfold emap_le in |- *.
      intro.
      rewrite <- Hcut.
      simpl in |- *.
      compare_cases x split_point.
     apply Hcorrect_l2.
     apply Hcorrect_r2.

  Qed.

  Definition pointwise_comp_widening :=
    (comp_widening_chain_intro (emap_set codomain) (Empty_emap codomain) 
    (fun y => Some (establish_widening y))).
*)
End widening.

Section Map1.
  Variable codomain1 codomain2 : Set.
  Variable f : codomain1 -> codomain2.

  Fixpoint emap_map1 (m : (emap_set codomain1))
    {struct m} : (emap_set codomain2) :=
    match m with
      Empty_emap => (Empty_emap codomain2)
    | (Node_emap l d c r h) =>
         (Node_emap codomain2
            (emap_map1 l)
            d (f c)
            (emap_map1 r)
            h)
    end.

  Lemma emap_map1_bounded_down :
    forall bound : domain,
    forall m : (emap_set codomain1),
    (emap_bounded_down codomain1 bound m) ->
    (emap_bounded_down codomain2 bound (emap_map1 m)).
  Proof.
    induction m; intro Hbound; inversion Hbound; simpl in |- *; auto with emap.
  Qed.

  Lemma emap_map1_bounded_up :
    forall bound : domain,
    forall m : (emap_set codomain1),
    (emap_bounded_up codomain1 bound m) ->
    (emap_bounded_up codomain2 bound (emap_map1 m)).
  Proof.
    induction m; intro Hbound; inversion Hbound; simpl in |- *; auto with emap.
  Qed.

  Hint Resolve emap_map1_bounded_down : emap.
  Hint Resolve emap_map1_bounded_up : emap.

  Lemma emap_map1_height :
    forall m : (emap_set codomain1),
    (height codomain2 (emap_map1 m)) = (height codomain1 m).
  Proof.
    induction m; auto with arith.
  Qed.

  Hint Rewrite emap_map1_height : emap.

  Lemma emap_map1_wo :
    forall m : (emap_set codomain1),
    (emap_wo codomain1 m) ->
    (emap_wo codomain2 (emap_map1 m)).
  Proof.
    induction m; intro Hwo; inversion Hwo; simpl in |- *; auto with emap.
    apply Node_emap_wo; auto with emap.
    rewrite emap_map1_height.
    rewrite emap_map1_height.
    assumption.
  Qed.

  Definition map1 : (t codomain1) -> (t codomain2).
  Proof.
    intro m.
    destruct m as (m, Hmwo).
    exists (emap_map1 m).
    apply emap_map1_wo; assumption.
  Defined.

  Definition func_map1 (mapf : domain -> (option codomain1))
    (lookup : domain) :=
    match mapf lookup with
    | None => None
    | Some x => Some (f x)
    end.

  Lemma emap_map1_correct :
    forall m : (emap_set codomain1),
    forall lookup : domain,
    (func_map1 (fun lookup1 => emap_find codomain1 lookup1 m)) lookup =
    emap_find codomain2 lookup (emap_map1 m).
  Proof.
    unfold func_map1 in |- *; induction m; simpl in |- *.
    auto.
    intro.
    compare_cases lookup d.
  Qed.

  Theorem map1_correct :
    forall m : (t codomain1),
    forall lookup : domain,
    (func_map1 (fun lookup1 => find codomain1 lookup1 m)) lookup =
    find codomain2 lookup (map1 m).
  Proof.
    destruct m; unfold find, map1 in |- *; simpl in |- *.
    apply emap_map1_correct.
  Qed.
End Map1.

Section Mapi1_opt.
  Variable codomain1 codomain2 : Set.
  Variable f : domain -> codomain1 -> (option codomain2).

  Fixpoint emap_mapi1_opt (m : (emap_set codomain1)) : (emap_set codomain2) :=
    match m with
    | Empty_emap => (Empty_emap codomain2)
    | (Node_emap l d c r _) =>
        let l2 := emap_mapi1_opt l in
        let r2 := emap_mapi1_opt r in
        match f d c with
        | None => merge codomain2 l2 r2
        | (Some c2) => create codomain2 l2 d c2 r2
        end
    end.

  Lemma emap_mapi1_opt_bounded_up:
    forall bound : domain,
    forall m : (emap_set codomain1),
    (emap_bounded_up codomain1 bound m) ->
    (emap_bounded_up codomain2 bound (emap_mapi1_opt m)).
  Proof.
    induction m; simpl in |- *; intros; inversions_all.
     auto with emap.
     destruct (f d c); auto with emap.
  Qed.

  Lemma emap_mapi1_opt_bounded_down:
    forall bound : domain,
    forall m : (emap_set codomain1),
    (emap_bounded_down codomain1 bound m) ->
    (emap_bounded_down codomain2 bound (emap_mapi1_opt m)).
  Proof.
    induction m; simpl in |- *; intros; inversions_all.
     auto with emap.
     destruct (f d c); auto with emap.
  Qed.

  Hint Resolve emap_mapi1_opt_bounded_up emap_mapi1_opt_bounded_down : emap.

  Lemma emap_mapi1_opt_wo:
    forall m : (emap_set codomain1),
    (emap_wo codomain1 m) ->
    (emap_wo codomain2 (emap_mapi1_opt m)).
  Proof.
    induction m; simpl in |- *; intros; inversions_all.
     auto with emap.
     destruct (f d c).
      auto with emap.
      apply merge_wo with (middle := d); auto with emap.
  Qed.

  Hint Resolve emap_mapi1_opt_wo : emap.

  Definition func_mapi1_opt (m : domain -> (option codomain1))
    (lookup : domain) : (option codomain2) :=
    match (m lookup) with
    | None => None
    | (Some data) => f lookup data
    end.

  Hypothesis f_stable :
    forall d1 d2 : domain,
    (domain_le d1 d2) -> (domain_le d2 d1) ->
    forall c : codomain1,
    (f d1 c) = (f d2 c).

  Lemma emap_mapi1_opt_correct :
    forall m : (emap_set codomain1),
    (emap_wo codomain1 m) ->
    forall lookup : domain,
    (emap_find codomain2 lookup (emap_mapi1_opt m)) =
    (func_mapi1_opt (fun lookup1 => emap_find codomain1 lookup1 m) lookup).
  Proof.
    unfold func_mapi1_opt in |- *.
    induction m; simpl in |- *; intros; inversions_all.
     trivial.
     compare_cases lookup d.
      rewrite <- IHm1; trivial.
    destruct (f d c); simpl in |- *.
       order_rewrites; trivial.
       apply merge_correct1 with (middle := d); auto with emap.
      rewrite (f_stable lookup d); trivial.
    destruct (f d c).
       simpl in |- *; order_rewrites; auto with emap.
       apply merge_correctM with (middle := d); auto with emap.
      rewrite <- IHm2; trivial.
    destruct (f d c); simpl in |- *.
       order_rewrites; trivial.
       apply merge_correct2 with (middle := d); auto with emap.
  Qed.

  Hint Resolve emap_mapi1_opt_correct : emap.

  Definition mapi1_opt (m : (t codomain1)) : (t codomain2).
  Proof.
    intro.
    destruct m as (m, Hm).
    exists (emap_mapi1_opt m).
    auto with emap.
  Defined.

  Theorem mapi1_opt_correct :
    forall m : (t codomain1),
    forall lookup : domain,
    (find codomain2 lookup (mapi1_opt m)) =
    (func_mapi1_opt (fun lookup1 => find codomain1 lookup1 m) lookup).
  Proof.
    unfold find, mapi1_opt in |- *.
    destruct m as (m, Hm); simpl in |- *.
    auto with emap.
  Qed.
End Mapi1_opt.

Hint Resolve emap_mapi1_opt_bounded_up emap_mapi1_opt_bounded_down emap_mapi1_opt_wo emap_mapi1_opt_correct : emap.

Section Mapi2_opt.
  Variable codomain1 codomain2 codomain3 : Set.
  Variable f1 : domain -> codomain1 -> (option codomain3).
  Variable f2 : domain -> codomain2 -> (option codomain3).
  Variable g : domain -> codomain1 -> codomain2 -> (option codomain3).

  Fixpoint emap_mapi2_opt (m1 : (emap_set codomain1))
    (m2 : (emap_set codomain2)) { struct m1 } : (emap_set codomain3) :=
    match m1 with
    | Empty_emap => (emap_mapi1_opt codomain2 codomain3 f2 m2)
    | (Node_emap l1 d c1 r1 _) =>
      let (l2, c2o, r2) :=  (cut codomain2 d m2) in
      let l := emap_mapi2_opt l1 l2 in
      let r := emap_mapi2_opt r1 r2 in
      match
        match c2o with
        | Some c2 => g d c1 c2
        | None => f1 d c1
        end with
      | None => merge codomain3 l r
      | Some c => create codomain3 l d c r
      end
    end.
      
  Lemma emap_mapi2_opt_bounded_up:
    forall bound : domain,
    forall m1 : (emap_set codomain1),
    forall m2 : (emap_set codomain2),
    (emap_bounded_up codomain1 bound m1) ->
    (emap_bounded_up codomain2 bound m2) ->
    (emap_bounded_up codomain3 bound (emap_mapi2_opt m1 m2)).
  Proof.
    induction m1; simpl in |- *; intros.
     auto with emap.
     Make.inversions_all.
       assert (cut_record_bounded_up codomain2 bound (cut codomain2 d m2)).
      apply cut_bounded_up; trivial.
      destruct (cut codomain2 d m2) as (l2, c2o, r2).
    inversions_all.
    destruct c2o as [c2| ].
       destruct (g d c c2); auto with emap.
       destruct (f1 d c); auto with emap.
  Qed.

  Lemma emap_mapi2_opt_bounded_down:
    forall bound : domain,
    forall m1 : (emap_set codomain1),
    forall m2 : (emap_set codomain2),
    (emap_bounded_down codomain1 bound m1) ->
    (emap_bounded_down codomain2 bound m2) ->
    (emap_bounded_down codomain3 bound (emap_mapi2_opt m1 m2)).
  Proof.
    induction m1; simpl in |- *; intros.
     auto with emap.
     Make.inversions_all.
       assert (cut_record_bounded_down codomain2 bound (cut codomain2 d m2)).
      apply cut_bounded_down; trivial.
      destruct (cut codomain2 d m2) as (l2, c2o, r2).
    inversions_all.
    destruct c2o as [c2| ].
       destruct (g d c c2); auto with emap.
       destruct (f1 d c); auto with emap.
  Qed.

  Hint Resolve emap_mapi2_opt_bounded_up emap_mapi2_opt_bounded_down : emap.
      
  Lemma emap_mapi2_opt_wo:
    forall m1 : (emap_set codomain1),
    forall m2 : (emap_set codomain2),
    (emap_wo codomain1 m1) -> (emap_wo codomain2 m2) ->
    (emap_wo codomain3 (emap_mapi2_opt m1 m2)).
  Proof.
    induction m1; simpl in |- *; intros.
     auto with emap.
     inversions_all.
       assert (cut_record_wo codomain2 d (cut codomain2 d m2)).
      apply cut_wo; trivial.
      destruct (cut codomain2 d m2) as (l2, c2o, r2).
    inversions_all.
    destruct c2o as [c2| ].
       destruct (g d c c2).
    auto with emap.
    apply merge_wo with (middle := d); auto with emap.
       destruct (f1 d c).
    auto with emap.
    apply merge_wo with (middle := d); auto with emap.
  Qed.

  Hint Resolve emap_mapi2_opt_wo : emap.

  Definition func_mapi2_opt
    (m1 : domain -> (option codomain1))
    (m2 : domain -> (option codomain2))
    (lookup : domain) : (option codomain3) :=
    match (m1 lookup) with
    | (Some c1) =>
      match (m2 lookup) with
      | (Some c2) => (g lookup c1 c2)
      | None => (f1 lookup c1)
      end
    | None =>
      match (m2 lookup) with
      | (Some c2) => (f2 lookup c2)
      | None => None
      end
    end.

  Hypothesis f1_stable :
    forall d1 d2 : domain,
    (domain_le d1 d2) -> (domain_le d2 d1) ->
    forall c : codomain1,
    (f1 d1 c) = (f1 d2 c).

  Hypothesis f2_stable :
    forall d1 d2 : domain,
    (domain_le d1 d2) -> (domain_le d2 d1) ->
    forall c : codomain2,
    (f2 d1 c) = (f2 d2 c).

  Hypothesis g_stable :
    forall d1 d2 : domain,
    (domain_le d1 d2) -> (domain_le d2 d1) ->
    forall c1 : codomain1,
    forall c2 : codomain2,
    (g d1 c1 c2) = (g d2 c1 c2).

  Lemma emap_mapi2_opt_correct :
    forall m1 : (emap_set codomain1),
    forall m2 : (emap_set codomain2),
    (emap_wo codomain1 m1) -> (emap_wo codomain2 m2) ->
    forall lookup : domain,
    (emap_find codomain3 lookup (emap_mapi2_opt m1 m2)) =
    (func_mapi2_opt
      (fun lookup1 => (emap_find codomain1 lookup1 m1))
      (fun lookup1 => (emap_find codomain2 lookup1 m2)) lookup).
  Proof.
    unfold func_mapi2_opt in |- *.
    induction m1; simpl in |- *; intros.
     rewrite emap_mapi1_opt_correct; trivial.
     inversions_all.
       assert (Hwo : cut_record_wo codomain2 d (cut codomain2 d m2)).
      auto with emap.
      assert
       (rew :
    cut_find codomain2 d lookup (cut codomain2 d m2) =
    emap_find codomain2 lookup m2).
       apply cut_correct; trivial.
       destruct (cut codomain2 d m2) as (l2, c2o, r2).
     simpl in rew.
     inversion_clear Hwo.
     destruct c2o as [c2| ].
    compare_cases lookup d.
     destruct (g d c c2) as [c3| ]; simpl in |- *; rewrite <- rew;
      order_rewrites.
      rewrite IHm1_1; trivial.
      rewrite merge_correct1 with (middle := d); auto with emap.
     rewrite (g_stable d lookup); trivial.
       rewrite <- rew.
       destruct (g lookup c c2) as [c3| ]; simpl in |- *; order_rewrites.
      trivial.
      apply merge_correctM with (middle := d); auto with emap.
     destruct (g d c c2) as [c3| ]; simpl in |- *; rewrite <- rew;
      order_rewrites.
      rewrite IHm1_2; trivial.
      rewrite merge_correct2 with (middle := d); auto with emap.
    compare_cases lookup d.
     destruct (f1 d c) as [c3| ]; simpl in |- *; order_rewrites.
      rewrite IHm1_1; trivial.
        rewrite <- rew.
        trivial.
      rewrite merge_correct1 with (middle := d); auto with emap.
        rewrite IHm1_1; trivial.
        rewrite <- rew.
        trivial.
     rewrite (f1_stable d lookup); trivial.
       destruct (f1 lookup c) as [c3| ].
      simpl in |- *; order_rewrites.
        rewrite <- rew.
        trivial.
      rewrite <- rew.
        apply merge_correctM with (middle := d); auto with emap.
     destruct (f1 d c) as [c3| ]; simpl in |- *; order_rewrites.
      rewrite IHm1_2; trivial.
        rewrite <- rew.
        trivial.
      rewrite merge_correct2 with (middle := d); auto with emap.
        rewrite IHm1_2; trivial.
        rewrite <- rew.
        trivial.
  Qed.

  Definition mapi2_opt
    (m1 : (t codomain1)) (m2 : (t codomain2)) : (t codomain3).
  Proof.
    intros.
    destruct m1 as (m1, Hm1).
    destruct m2 as (m2, Hm2).
    exists (emap_mapi2_opt m1 m2).
    auto with emap.
  Defined.

  Theorem mapi2_opt_correct:
    forall m1 : (t codomain1),
    forall m2 : (t codomain2),
    forall lookup : domain,
    (find codomain3 lookup (mapi2_opt m1 m2)) =
    (func_mapi2_opt
      (fun lookup1 => (find codomain1 lookup1 m1))
      (fun lookup1 => (find codomain2 lookup1 m2)) lookup).
  Proof.
    intros.
    destruct m1 as (m1, Hm1).
    destruct m2 as (m2, Hm2).
    unfold find, mapi2_opt in |- *.
    simpl in |- *.
    apply emap_mapi2_opt_correct; auto with emap.
  Qed.
End Mapi2_opt.
End Make.
